# -*- coding: utf-8 -*-
"""
Module containing the parser function that can read an XML Schema file and
return an :class:`~pyxst.xml_struct.xsd.root.XsdSchema` object.

.. autofunction:: parse_schema
"""

from .errors import XsdError
from xml.parsers.expat import ExpatError
from traceback import format_exc
from os.path import abspath, dirname, join
from lxml.etree import parse
from .root import XsdSchema, CAT_MAPPING
from .types import XsdComplexType
from .utils import XSD_NS
from .utils import split_qname


def parse_schema(xsd_file, sch_obj=None, target_ns=None,
                 ns_strict_equality=False):
    """
    Reads the ``xsd_file`` and returns an XsdSchema object containing
    the nodes (cf. nodes derived from
    :class:`~pyxst.xml_struct.xsd.utils.XsdNode`) read in
    the XML Schema.

    These nodes represent all the information read inside the XSD Schema
    (definition of XML elements and XML attributes, declaration of types,
    groups and attribute groups).

    :param xsd_file:
        Name of the file that contains the XML Schema.
    :type xsd_file:
        :class:`str`
    :param sch_obj:
        Schema object where the nodes read from the XML Schema will be stored.
        If no schema object is provided, a new one is created.
    :type sch_obj:
        :class:`~pyxst.xml_struct.xsd.root.XsdSchema`
    :param target_ns:
        Target namespace used to qualify the names of the nodes read in the
        XML Schema. If a namespace is specified in the XML Schema, it should
        be equal to the namespace specified in this parameter.
    :type target_ns:
        :class:`unicode`
    :param ns_strict_equality:
        If set to ``True``, the namespace specified in the XML Schema
        and the namespace specified in ``target_ns`` parameter must be
        strictly equals: either they are both ``None`` or they both
        have the same value. If set to ``False``, if the namespace
        specified in ``target_ns`` parameter is ``None``, it will be
        replaced by the namespace specified in the XML Schema (but if
        they both have a value, these values must be the same).
    :type ns_strict_equality:
        :class:`boolean`
    :returns:
        An XsdSchema object grouping all the nodes read from the XML Schema.
    :rtype:
        :class:`~pyxst.xml_struct.xsd.root.XsdSchema`
    """
    # Opens the XML Schema file
    try:
        xsd_file = abspath(xsd_file)
    except Exception:
        pass
    try:
        xml_tree = parse(xsd_file)
    except IOError:
        raise XsdError("Can't open the %s XML Schema file\n%s"
                       % (xsd_file, format_exc()))
    except ExpatError:
        raise XsdError("Found a syntax error while reading the %s XML "
                       "Schema file\n%s" % (xsd_file, format_exc()))
    xml_root = xml_tree.getroot()
    if xml_root.tag != u"{%s}schema" % XSD_NS:
        raise XsdError("%s XML file doesn't contain an XML Schema"
                       % xsd_file)
    if sch_obj is None:
        sch_obj = XsdSchema()
        sch_obj.loaded_files = set([xsd_file])
    else:
        sch_obj.loaded_files.add(xsd_file)
    # Checks target namespace
    if not ns_strict_equality:
        if target_ns is None:
            target_ns = xml_root.get(u"targetNamespace")
        elif xml_root.get(u"targetNamespace") is not None and \
                target_ns != xml_root.get(u"targetNamespace"):
            raise XsdError("Target namespace of %s should be %s instead of %s"
                           % (xsd_file, target_ns,
                              xml_root.get(u"targetNamespace")))
    elif target_ns != xml_root.get(u"targetNamespace"):
        raise XsdError("Target namespace of %s should be %s instead of %s"
                       % (xsd_file, target_ns,
                          xml_root.get(u"targetNamespace")))
    # Processes inclusions or imports of external XML Schemas
    for xml_obj in xml_root.findall(u"{%s}include" % XSD_NS):
        ext_file = xml_obj.get(u"schemaLocation")
        ext_file = abspath(join(dirname(xsd_file), ext_file))
        if not ext_file in sch_obj.loaded_files:
            parse_schema(ext_file, sch_obj, target_ns)
    for xml_obj in xml_root.findall(u"{%s}import" % XSD_NS):
        ext_file = xml_obj.get(u"schemaLocation")
        ext_file = abspath(join(dirname(xsd_file), ext_file))
        if not ext_file in sch_obj.loaded_files:
            ext_ns = xml_obj.get(u"namespace")
            parse_schema(ext_file, sch_obj, ext_ns, ns_strict_equality=True)
    # Processes redefinitions from external XML Schemas
    for xml_obj in xml_root.findall(u"{%s}redefine" % XSD_NS):
        ext_file = xml_obj.get(u"schemaLocation")
        ext_file = join(dirname(xsd_file), ext_file)
        raise XsdError("<redefine> objects are not yet supported "
                       "(%s XML Schema)" % xsd_file)
    # Sets global properties in the schema object
    sch_obj._xsd_file = xsd_file
    sch_obj._target_ns = target_ns
    sch_obj._qualified_elts = (xml_root.get(u"elementFormDefault") ==
                               u"qualified")
    sch_obj._qualified_attrs = (xml_root.get(u"attributeFormDefault") ==
                                u"qualified")
    # Firstly, creates the high level nodes defined in the XML Schema
    for xml_obj in xml_root:
        ns, obj_sort = split_qname(xml_obj.tag)
        if ns == XSD_NS and obj_sort in CAT_MAPPING:
            node = sch_obj.create_node(xml_obj)
            sch_obj.store_definition_node(node)
    # Then, fills the nodes with the information defined in the XML Schema
    # (starts with high level nodes and goes down in hierarchy)
    filling_function = {u"element"       : sch_obj.fill_element_node,
                        u"attributeGroup": sch_obj.fill_attribute_group_node,
                        u"attribute"     : sch_obj.fill_attribute_node,
                        u"complexType"   : sch_obj.fill_complex_type_node,
                        u"simpleType"    : sch_obj.fill_simple_type_node,
                        u"group"         : sch_obj.fill_group_node,
                        u"all"           : sch_obj.fill_group_node,
                        u"choice"        : sch_obj.fill_group_node,
                        u"sequence"      : sch_obj.fill_group_node,
                        }
    for xml_obj in xml_root:
        ns, obj_sort = split_qname(xml_obj.tag)
        if ns == XSD_NS and obj_sort in CAT_MAPPING:
            name = sch_obj.get_node_name(xml_obj)
            node = sch_obj.get_definition_node(CAT_MAPPING[obj_sort], name)
            filling_function[obj_sort](node, xml_obj)
    # Resolves the attribute groups and the derivation in the types
    for typ in sch_obj.all_types:
        typ.resolve_derivation()
        if isinstance(typ, XsdComplexType):
            typ.resolve_attribute_groups()
    # Unsets the global properties from the schema object as reading is finished
    sch_obj._xsd_file = ""
    sch_obj._target_ns = None
    sch_obj._qualified_elts = False
    sch_obj._qualified_attrs = False
    return sch_obj



# -*- coding: utf-8 -*-
"""
This module defines the class where the data read from an XML Schema will
be stored.

Hence, the data read from an XML Schema is modeled as nodes deriving
from :class:`~pyxst.xml_struct.xsd.utils.XsdNode` and
these nodes are stored in a dedicated structure: the
:class:`~pyxst.xml_struct.xsd_reader.XsdSchema` class.

This class also has methods to build the nodes from the XML Schema data.

.. autodata:: XSD_BUILT_IN_TYPES

.. autodata:: CAT_MAPPING

.. autodata:: NODE_MAPPING

.. autoclass:: XsdSchema
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"


from .errors import XsdError
from .utils import (
    ATTRIBUTE, ELEMENT, GROUP, ATTRIBUTE_GROUP, TYPE, XSD_NS, XsdUsage)
from .utils import (INFINITY, MIXED, split_qname, resolve_xml_qname)
from .elements import XsdElement
from .attributes import (XsdAttribute, XsdAttributeGroup)
from .types import (
    XsdComplexType, XsdSimpleType, XsdSimpleListType, XsdSimpleUnionType,
    XsdType, CONSTRAINT_MAPPING, RESTRICTION, EXTENSION)
from .groups import (XsdChoice, XsdAll, XsdSequence)
from traceback import format_exc
from lxml import etree


XSD_BUILT_IN_TYPES = set([
        u"anyType", u"anySimpleType", u"boolean", u"base64Binary", u"hexBinary",
        u"float", u"decimal", u"integer", u"nonPositiveInteger",
        u"negativeInteger", u"long", u"int", u"short", u"byte",
        u"nonNegativeInteger", u"unsignedLong", u"unsignedInt",
        u"unsignedShort", u"unsignedByte", u"positiveInteger", u"double",
        u"anyURI", u"QName", u"NOTATION", u"duration", u"dateTime", u"time",
        u"date", u"gYearMonth", u"gYear", u"gMonthDay", u"gDay", u"gMonth",
        u"string", u"normalizedString", u"token", u"language", u"Name",
        u"NCName", u"ID", u"IDREF", u"IDREFS", u"ENTITY", u"ENTITIES",
        u"NMTOKEN", u"NMTOKENS", ])
"""
Set containing the built-in types defined in the XSD standard.

Type: set of :class:`unicode`
"""

CAT_MAPPING = {u"element"                   : ELEMENT,
               u"attributeGroup"            : ATTRIBUTE_GROUP,
               u"attribute"                 : ATTRIBUTE,
               u"complexType"               : TYPE,
               u"simpleType"                : TYPE,
               u"simpleType/restriction"    : TYPE,
               u"simpleType/list"           : TYPE,
               u"simpleType/union"          : TYPE,
               u"group"                     : GROUP,
               u"all"                       : GROUP,
               u"choice"                    : GROUP,
               u"sequence"                  : GROUP,
               }
"""
Dictionary mapping the XML Schema item types (the name of the XML elements
inside the XML Schema) and the categories of nodes defined in
:mod:`~pyxst.xml_struct.xsd.utils` module.

Type: Dictionary of :class:`str` indexed with :class:`unicode`
"""

NODE_MAPPING = {u"element"                   : XsdElement,
                u"attributeGroup"            : XsdAttributeGroup,
                u"attribute"                 : XsdAttribute,
                u"complexType"               : XsdComplexType,
                u"simpleType"                : XsdSimpleType,
                u"simpleType/restriction"    : XsdSimpleType,
                u"simpleType/list"           : XsdSimpleListType,
                u"simpleType/union"          : XsdSimpleUnionType,
                u"all"                       : XsdAll,
                u"choice"                    : XsdChoice,
                u"sequence"                  : XsdSequence,
                }
"""
Dictionary mapping the XML Schema item types (the name of the XML elements
inside the XML Schema) and the classes of nodes defined in
:mod:`~pyxst.xml_struct.xsd.utils` module.

Type: Dictionary of :class:`class` indexed with :class:`unicode`
"""


class XsdSchema(object):
    """
    Class representing an XML Schema.

    This class stores the group nodes, the type nodes, the attribute
    group nodes, the attribute nodes and the element nodes defined at
    the highest level of the XML Schema.

    This class also has a factory method to build new nodes from XML objects
    read in the XML Schema.

    .. attribute:: groups

       Dictionary of the high level group nodes indexed with their name.

       Type: dictionary of
       :class:`~pyxst.xml_struct.xsd.groups.XsdGroup` indexed with
       :class:`unicode`

    .. attribute:: types

       Dictionary of the high level type nodes indexed with their name.

       Type: dictionary of
       :class:`~pyxst.xml_struct.xsd.types.XsdType` indexed with
       :class:`unicode`

    .. attribute:: attr_groups

       Dictionary of the high level attribute group nodes indexed with
       their name.

       Type: dictionary of
       :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`
       indexed with :class:`unicode`

    .. attribute:: attributes

       Dictionary of the high level attribute nodes indexed with their name.

       Type: dictionary of
       :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`
       indexed with :class:`unicode`

    .. attribute:: elements

       Dictionary of the high level element nodes indexed with their name.

       Type: dictionary of
       :class:`~pyxst.xml_struct.xsd.elements.XsdElement` indexed
       with :class:`unicode`

    .. attribute:: storages

       Dictionary indexing the previous attributes with the category of the
       nodes they will contain. For example, the ``elements`` attribute is
       indexed with the ``ELEMENT`` constant.

       Type: dictionary of :class:`dict` indexed with :class:`str`

    .. attribute:: all_types

       List containing all the types that exist in this schema object. The
       types include the high-level types (stored in ``types`` attribute)
       and the local types defined into other nodes.

       Type: List of
       :class:`~pyxst.xml_struct.xsd.types.XsdType`

    .. attribute:: _target_ns

       Private attribute containing the target namespace used to qualify
       the names of the items defined in the XML Schema that is currently
       read. If None, the names are left unqualified.

       This namespace changes when reading included XML Schemas.

       Type: :class:`unicode`

    .. attribute:: _qualified_attrs

       Private flag set to ``True`` if the names of the attributes in the
       schema are qualified names (namespace + local name). However, each
       attribute definition can locally change this behaviour.

       The value of this flag changes when reading included XML Schemas.

       Type: :class:`boolean`

    .. attribute:: _qualified_elts

       Private flag set to ``True`` if the names of the elements in the schema
       are qualified names (namespace + local name). However, each
       element definition can locally change this behaviour.

       The value of this flag changes when reading included XML Schemas.

       Type: :class:`boolean`

    .. attribute:: _xsd_file

       Private attribute containing the file name of the XML Schema that is
       currently read. It is mainly use when raising exceptions.

       This name changes when reading included XML Schemas.

       Type: :class:`str`

    .. automethod:: __init__
    """
    def __init__(self):
        """
        Initializes a new XsdSchema root node.
        """
        self.groups = {}
        self.types = {}
        self.attr_groups = {}
        self.attributes = {}
        self.elements = {}
        self.storages = {ATTRIBUTE: self.attributes,
                         ATTRIBUTE_GROUP: self.attr_groups,
                         ELEMENT: self.elements,
                         TYPE: self.types,
                         GROUP: self.groups, }
        self.all_types = []
        self._target_ns = None
        self._qualified_attrs = False
        self._qualified_elts = False
        self._xsd_file = ""
        # Initializes the most basic built-in types (anyType and anySimpleType)
        for local_name in (u"anyType", u"anySimpleType"):
            type_name = u"{{{0}}}{1}".format(XSD_NS, local_name)
            type_node = XsdType(type_name)
            self.store_definition_node(type_node)

    def store_definition_node(self, node):
        """
        Stores a new high level node in this object.

        Checks there is not already a node with the same name in the
        corresponding storage.

        :param node:
            Node to be stored in this object.
        :type item_type:
            :class:`~pyxst.xml_struct.xsd.utils.XsdNode`
        """
        stor = self.storages[node.category]
        if node.name in stor:
            raise XsdError("Found a second definition for %s %s in the %s XML "
                           "Schema"
                           % (node.category, node.name, self._xsd_file))
        stor[node.name] = node
        if node.category is TYPE:
            self.all_types.append(node)

    def get_definition_node(self, category, name):
        """
        Returns the high level definition node with the ``name`` name
        extracted from the storage corresponding to ``category``.

        This node has been previously stored in one of the storages of this
        class. If it can't be found, raises an error.

        :param category:
            Category (element, attribute, attribute group, type, group) of
            the node to be found.
        :type item_type:
            :class:`str`
        :param name:
            Name of the node to be found.
        :type name:
            :class:`unicode`
        :returns:
            The definition node.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.utils.XsdNode`
        """
        try:
            stor = self.storages[category]
        except KeyError:
            raise XsdError("Can't find node of unknown category (%s)"
                           % category)
        if name not in stor:
            raise XsdError("Can't find the definition node for %s %s declared "
                           "in the %s XML Schema"
                           % (category, name, self._xsd_file))
        return stor[name]

    def create_node(self, xml_obj, name=None, uid_prefix=u""):
        """
        Creates a new node from the information extracted from ``xml_obj``
        XML element read in an XML Schema.

        The created node is empty: it is just an object of the correct class
        with its name set.

        The created node is not added to the storages where only high level
        nodes are kept. If the node must be kept in such a storage, the
        ``store_definition_node`` must be called.

        :param xml_obj:
            XML element defining a node (element, attribute, type, etc.) in
            an XML Schema.
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :param name:
            Name specified for the new node. If left to ``None``, the name is
            extracted from ``xml_obj``. In this case, if no name is defined in
            ``xml_obj`` (in ``name`` attribute), an exception is raised.
        :type name:
            :class:`unicode`
        :returns:
            A new node corresponding to ``xml_obj``
        :rtype:
            :class:`~pyxst.xml_struct.xsd.utils.XsdNode`
        """
        # Extracts the name of the node
        if name is None:
            name = self.get_node_name(xml_obj)
        if name is None:
            raise XsdError("Found an object (%s) without a name in %s XML "
                           "Schema" % (xml_obj, self._xsd_file))
        obj_ns, obj_sort = split_qname(xml_obj.tag)
        # Deals with specific sorts of XML Schema objects
        if obj_sort == u"group":
            # Finds the child node that is not an annotation to know what
            # kind of item this will be (choice, sequence or all)
            for obj_child in xml_obj:
                ns, obj_child_sort = split_qname(obj_child.tag)
                if ns != XSD_NS or obj_child_sort == u"annotation":
                    continue
                obj_sort = obj_child_sort
        if obj_sort == u"simpleType":
            # Finds the child node that is not an annotation to know what
            # kind of simple type this will be (simple type, simple list type
            # or simple union type)
            for obj_child in xml_obj:
                ns, obj_child_sort = split_qname(obj_child.tag)
                if ns != XSD_NS or obj_child_sort == u"annotation":
                    continue
                obj_sort = u"simpleType/" + obj_child_sort
        if obj_ns != XSD_NS or obj_sort not in NODE_MAPPING:
            raise XsdError("Can't create a node from %s object in %s XML "
                           "Schema" % (xml_obj, self._xsd_file))
        return NODE_MAPPING[obj_sort](name, uid_prefix=uid_prefix)

    def get_node_name(self, xml_obj):
        """
        Gets the name of the node that corresponds to the ``xml_obj``
        XML element read in an XML Schema.

        Basically, this name is the name defined in the ``name`` attribute of
        ``xml_obj`` but it is sometimes necessary to add a namespace before
        this local name.

        If ``xml_obj`` has no ``name`` attribute (anonymous object), ``None``
        is returned.

        :param xml_obj:
            XML element defining a node (element, attribute, type, etc.) in
            an XML Schema.
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            The name of the node that corresponds to ``xml_obj``
        :rtype:
            :class:`unicode`
        """
        _, def_categ = split_qname(xml_obj.tag)
        name = xml_obj.get(u"name")
        # No name specified, returns ``None``
        if name is None:
            return None
        # A namespace is specified, returns the qualified name
        if name.find(u":") != -1:
            return resolve_xml_qname(name, xml_obj.nsmap)
        # No namespace is specified, adds the target namespace if necessary
        add_target_ns = False
        if self._target_ns is not None:
            if xml_obj.getparent().tag != u"{{{0}}}schema".format(XSD_NS):
                # This is not a top-level definition
                if CAT_MAPPING[def_categ] is ATTRIBUTE:
                    if xml_obj.get(u"form") is None:
                        add_target_ns = self._qualified_attrs
                    else:
                        add_target_ns = (xml_obj.get(u"form") == u"qualified")
                elif CAT_MAPPING[def_categ] is ELEMENT:
                    if xml_obj.get(u"form") is None:
                        add_target_ns = self._qualified_elts
                    else:
                        add_target_ns = (xml_obj.get(u"form") == u"qualified")
            else:
                # Always adds target namespace to top-level definition and
                # definition that are neither elements nor attributes.
                add_target_ns = True
        if add_target_ns:
            return u"{{{0}}}{1}".format(self._target_ns, name)
        else:
            return name

    def get_type_nodes_from_xml(self, xml_obj, ref_attr_name, uid_prefix=u""):
        """
        Gets the type nodes that are referenced or contained into
        the ``xml_obj`` XML element read in an XML Schema.

        The type nodes can either be high level types referenced in
        the ``ref_attr_name`` attribute of ``xml_obj`` or new local types
        directly defined in children of ``xml_obj``. These children are
        ``simpleType`` or ``complexType`` XML elements.

        Most of the times, only one type node is referenced or
        defined but in the case of an ``union`` XML element, several
        type nodes are referenced or defined.

        :param xml_obj:
            XML element containing or referencing a type in the XML
            Schema (typically, ``xsd:attribute``, ``xsd:simpleContent``,
            ``xsd:restriction``, ``xsd:element``, etc.)
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :param ref_attr_name:
            Name of the attribute of ``xml_obj`` that might contain
            references towards the types.
        :type ref_attr_name:
            :class:`unicode`
        :param name_prefix:
            Prefix added before the name created for anonymous types defined
            in the children of ``xml_obj``.
        :type name_prefix:
            :class:`unicode`
        :returns:
            The list of type nodes referenced or defined into ``xml_obj``
        :rtype:
            List of
            :class:`~pyxst.xml_struct.xsd.types.XsdType`
        """
        nodes = []
        ref_names = xml_obj.get(ref_attr_name, "")
        # First, gets the types specified with a reference
        for type_name in ref_names.split():
            type_name = resolve_xml_qname(type_name, xml_obj.nsmap)
            try:
                nodes.append(self.get_definition_node(TYPE, type_name))
            except XsdError:
                # If the referenced type is an unknown built-in type, adds
                # it as a high level type
                ns, type_local_name = split_qname(type_name)
                if ns == XSD_NS and type_local_name in XSD_BUILT_IN_TYPES:
                    type_node = XsdSimpleType(type_name)
                    self.store_definition_node(type_node)
                    nodes.append(type_node)
                else:
                    raise
        # Secondly, builds types defined in children of xml_obj
        s_num = c_num = 0
        for child_obj in xml_obj:
            ns, child_name = split_qname(child_obj.tag)
            if ns == XSD_NS and child_name == u"simpleType":
                type_name = "simpleType%d" % s_num
                type_node = self.create_node(child_obj, name=type_name,
                                             uid_prefix=uid_prefix)
                self.fill_simple_type_node(type_node, child_obj)
                s_num += 1
                self.all_types.append(type_node)
                nodes.append(type_node)
            if ns == XSD_NS and child_name == u"complexType":
                type_name = "complexType%d" % c_num
                type_node = self.create_node(child_obj, name=type_name,
                                             uid_prefix=uid_prefix)
                self.fill_complex_type_node(type_node, child_obj)
                c_num += 1
                self.all_types.append(type_node)
                nodes.append(type_node)
        # Thirdly, if no type has been found and xml_obj is an element or
        # an attribute, uses "anyType" or "anySimpleType" as its type
        ns, loc_name = split_qname(xml_obj.tag)
        if len(nodes) == 0 and loc_name == "element":
            nodes.append(self.get_definition_node(
                TYPE, u"{{{0}}}anyType".format(XSD_NS)))
        elif len(nodes) == 0 and loc_name == "attribute":
            nodes.append(self.get_definition_node(
                TYPE, u"{{{0}}}anySimpleType".format(XSD_NS)))
        return nodes

    def get_group_node_from_xml(self, xml_obj, name="", uid_prefix=u""):
        """
        Gets the group node that corresponds to the ``xml_obj``
        XML element read in an XML Schema.

        The group node can either be a high level group referenced in
        the ``ref`` attribute of ``xml_obj`` if this element is a
        ``group`` XML element or a local group defined in ``xml_obj``
        if this element is a ``all``, a ``choice`` or a ``sequence``
        XML elements

        :param xml_obj:
            XML element corresponding to a group in the XML
            Schema (typically, ``xsd:group``, ``xsd:all``, etc.)
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :param name:
            Name of the group if the group is defined in ``xml_obj``.
        :type name:
            :class:`unicode`
        :returns:
            The group node that corresponds to ``xml_obj``
        :rtype:
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        """
        if xml_obj.tag == "{%s}group" % XSD_NS:
            # Group is specified with a reference
            ref_name = resolve_xml_qname(xml_obj.get(u"ref"), xml_obj.nsmap)
            if ref_name is None:
                raise XsdError("Group %s should have a ref attribute to point "
                               "towards a defined group in %s XML Schema"
                               % (xml_obj, self._xsd_file))
            node = self.get_definition_node(GROUP, ref_name)
        else:
            node = self.create_node(xml_obj, name=name, uid_prefix=uid_prefix)
            self.fill_group_node(node, xml_obj)
        return node

    def get_element_node_from_xml(self, xml_obj, uid_prefix=u""):
        """
        Gets the element node that corresponds to the ``xml_obj``
        XML element read in an XML Schema.

        The element node can either be a high level element referenced with
        a ``ref`` attribute in ``xml_obj`` or a new local element directly
        defined by ``xml_obj``.

        :param xml_obj:
            XML element defining an element use in the XML Schema (typically,
            ``xsd:element``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            The element node that corresponds to ``xml_obj``
        :rtype:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdElement`
        """
        ref_name = resolve_xml_qname(xml_obj.get(u"ref"), xml_obj.nsmap)
        if ref_name is not None:
            # Element is specified with a reference
            node = self.get_definition_node(ELEMENT, ref_name)
        else:
            # Element is defined in xml_obj
            node = self.create_node(xml_obj, uid_prefix=uid_prefix)
            self.fill_element_node(node, xml_obj)
        return node

    def get_attribute_node_from_xml(self, xml_obj, uid_prefix=u""):
        """
        Gets the attribute node that corresponds to the ``xml_obj``
        XML element read in an XML Schema.

        The attribute node can either be a high level attribute referenced with
        a ``ref`` attribute in ``xml_obj`` or a new local attribute directly
        defined by ``xml_obj``.

        :param xml_obj:
            XML element defining an attribute use in the XML Schema (typically,
            ``xsd:attribute``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            The attribute node that corresponds to ``xml_obj``
        :rtype:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`
        """
        ref_name = resolve_xml_qname(xml_obj.get(u"ref"), xml_obj.nsmap)
        if ref_name is not None:
            # Attribute is specified with a reference
            node = self.get_definition_node(ATTRIBUTE, ref_name)
        else:
            # Attribute is defined in xml_obj
            node = self.create_node(xml_obj, uid_prefix=uid_prefix)
            self.fill_attribute_node(node, xml_obj)
        return node

    def get_min_max_occur_from_xml(self, xml_obj, item_node):
        """
        Returns the minimum and maximum occurences specified on ``xml_obj``
        for ``item_node``.

        If they are not specified, the default value (``1``) is returned.

        Minimum and maximum occurences will be used to insert the ``item_node``
        inside its parent node.

        :param xml_obj:
            XML element defining the use of an element or a group in the XML
            Schema (typically, ``xsd:element``, ``xsd:group``, ``xsd:sequence``,
            etc.)
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :param item_node:
            Node corresponding to ``xml_obj``. This node is just used to
            display context information in case of error.
        :type item_node:
            :class:`~pyxst.xml_struct.xsd.utils.XsdNode`
        :returns:
            Tuple composed with the minimum occurence and the maximum occurence
            of ``item_node``. The maximum occurence can be either an integer
            or ``INFINITY``.
        :rtype:
            tuple composed with :class:`int`, (:class:`int` or
            :data:`~pyxst.xml_struct.utils.INFINITY`)
        """
        if xml_obj.get(u"maxOccurs") == u"unbounded":
            max_occur = INFINITY
        else:
            try:
                max_occur = int(xml_obj.get(u"maxOccurs", 1))
            except ValueError:
                raise XsdError("Can't read maximum occurence for %s "
                               "(should be an integer) in %s XML Schema:\n%s"
                               % (item_node, self._xsd_file, format_exc()))
        try:
            min_occur = int(xml_obj.get(u"minOccurs", 1))
        except ValueError:
            raise XsdError("Can't read minimum occurence for %s "
                           "(should be an integer) in %s XML Schema:\n%s"
                           % (item_node, self._xsd_file, format_exc()))
        return min_occur, max_occur

    def add_attributes_to_node(self, node, xml_obj):
        """
        Adds into the ``node`` object the attributes and attribute groups
        found in the ``xml_obj`` XML element read in an XML Schema.

        :param node:
            Attribute group node or Complex type node where attributes and
            attribute sub-groups must be added.
        :param node:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`
            or :class:`~pyxst.xml_struct.xsd.type.XsdComplexType`
        :param xml_obj:
            XML element containing definitions of attributes and attribute
            groups in the XML Schema (typically, ``xsd:attributeGroup``,
            ``xsd:complexType``, ``xsd:complexContent``, ``xsd:simpleContent``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after the attributes and attribute groups have been added.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`
            or :class:`~pyxst.xml_struct.xsd.type.XsdComplexType`
        """
        for child_obj in xml_obj:
            ns, child_name = split_qname(child_obj.tag)
            if ns == XSD_NS and child_name == u"attributeGroup":
                ref_name = resolve_xml_qname(child_obj.get(u"ref"),
                                             child_obj.nsmap)
                grp_node = self.get_definition_node(ATTRIBUTE_GROUP,
                                                    ref_name)
                node.add_attribute_group(grp_node)
            elif ns == XSD_NS and child_name == u"attribute":
                attr_node = self.get_attribute_node_from_xml(
                    child_obj, uid_prefix=node.uid)
                attr_use = child_obj.get(u"use")
                if attr_use == u"prohibited":
                    min_occur = 0
                    max_occur = 0
                elif attr_use == u"required":
                    min_occur = 1
                    max_occur = 1
                else: # u"optional" is the default value
                    min_occur = 0
                    max_occur = 1
                node.add_attribute(attr_node, min_occur, max_occur)
        return node

    def fill_element_node(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Element node that must be filled with the information
            available in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement`
        :param xml_obj:
            XML element defining an element in the XML Schema
            (typically, ``xsd:element``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement`
        """
        self.collect_documentation(node, xml_obj)
        type_nodes = self.get_type_nodes_from_xml(xml_obj, u"type",
                                                  uid_prefix=node.uid)
        if len(type_nodes) == 0:
            # Tries to get the type from the substitution group
            subs_name = xml_obj.get("substitutionGroup")
            if subs_name is not None:
                subs_node = self.get_definition_node(ELEMENT, subs_name)
                if subs_node.xsd_type is not None:
                    node.xsd_type = subs_node.xsd_type
                else:
                    subs_node._elts_with_same_type.append(node)
            else:
                raise XsdError("%s should contain either a reference or an "
                               "inner definition of type in %s XML Schema"
                               % (node, self._xsd_file))
        node.xsd_type = type_nodes[0]
        return node

    def fill_attribute_group_node(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Attribute group node that must be filled with the information
            available in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`
        :param xml_obj:
            XML element defining an attribute group in the XML Schema
            (typically, ``xsd:attributeGroup``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`
        """
        self.collect_documentation(node, xml_obj)
        self.add_attributes_to_node(node, xml_obj)
        return node

    def fill_attribute_node(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        Filling an attribute only consists in linking it to its type node.

        :param node:
            Attribute node that must be filled with the information available
            in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`
        :param xml_obj:
            XML element defining an attribute in the XML Schema (typically,
            ``xsd:attribute``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`
        """
        self.collect_documentation(node, xml_obj)
        type_nodes = self.get_type_nodes_from_xml(xml_obj, u"type",
                                                  uid_prefix=node.uid)
        if len(type_nodes) == 0:
            raise XsdError("%s should contain either a reference or an inner "
                           "definition of type in %s XML Schema"
                           % (node, self._xsd_file))
        node.xsd_type = type_nodes[0]
        return node

    def fill_simple_type_node(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Simple type node that must be filled with the information available
            in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.types.XsdSimpleType`
        :param xml_obj:
            XML element defining a simple type in the XML Schema (typically,
            ``xsd:simpleType``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.types.XsdSimpleType`
        """
        self.collect_documentation(node, xml_obj)
        if xml_obj.find(u"{%s}restriction" % XSD_NS) is not None:
            # Restriction of a base simple type
            rst_obj = xml_obj.find(u"{%s}restriction" % XSD_NS)
            self.collect_documentation(node, rst_obj)
            type_nodes = self.get_type_nodes_from_xml(
                rst_obj, u"base", uid_prefix=node.uid+u"/restriction")
            if len(type_nodes) == 0:
                raise XsdError("Restriction in %s should contain either a "
                               "reference or an inner definition of type in "
                               "%s XML Schema" % (node, self._xsd_file))
            node._base_type = type_nodes[0]
            node._deriv_method = RESTRICTION
            for child_obj in rst_obj:
                ns, child_name = split_qname(child_obj.tag)
                if ns != XSD_NS or child_name not in CONSTRAINT_MAPPING:
                    continue
                node.add_constraint(child_name, child_obj.get(u"value"))
        elif xml_obj.find(u"{%s}list" % XSD_NS) is not None:
            # List of several items (space separated) whose type is simple type
            lst_obj = xml_obj.find(u"{%s}list" % XSD_NS)
            type_nodes = self.get_type_nodes_from_xml(
                lst_obj, u"itemType", uid_prefix=node.uid+u"/list")
            if len(type_nodes) == 0:
                raise XsdError("List in %s should contain either a reference "
                               "or an inner definition of type in %s XML "
                               "Schema" % (node, self._xsd_file))
            node.listitem_type = type_nodes[0]
        elif xml_obj.find(u"{%s}union" % XSD_NS) is not None:
            # Union of several possible simple types
            uni_obj = xml_obj.find(u"{%s}union" % XSD_NS)
            type_nodes = self.get_type_nodes_from_xml(
                uni_obj, u"memberTypes", uid_prefix=node.uid+u"/union")
            if len(type_nodes) == 0:
                raise XsdError("Union in %s should contain references or "
                               "inner definitions of type in %s XML Schema"
                               % (node, self._xsd_file))
            for typ in type_nodes:
                node.possible_types.add(typ)
        else:
            raise XsdError("%s should contain either a <restriction> or a "
                           "<list> or an <union> child in the %s XML Schema"
                           % (node, self._xsd_file))
        return node

    def fill_complex_type_node(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Complex type node that must be filled with the information
            available in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.types.XsdComplexType`
        :param xml_obj:
            XML element defining a complex type in the XML Schema (typically,
            ``xsd:complexType``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.types.XsdComplexType`
        """
        self.collect_documentation(node, xml_obj)
        if xml_obj.find(u"{%s}simpleContent" % XSD_NS) is not None:
            self.fill_node_from_simple_content(node,
                                  xml_obj.find(u"{%s}simpleContent" % XSD_NS))
        elif xml_obj.find(u"{%s}complexContent" % XSD_NS) is not None:
            if xml_obj.get(u"mixed", u"false") == u"true":
                node.content_type = MIXED
            self.fill_node_from_complex_content(node,
                                 xml_obj.find(u"{%s}complexContent" % XSD_NS))
        else:
            if xml_obj.get(u"mixed", u"false") == u"true":
                node.content_type = MIXED
            grp_node = None
            for child_obj in xml_obj:
                ns, child_sort = split_qname(child_obj.tag)
                if ns == XSD_NS and child_sort == u"group":
                    grp_node = self.get_group_node_from_xml(child_obj,
                                                            uid_prefix=node.uid)
                    break
                if ns == XSD_NS and child_sort in [u"all", u"choice",
                                                   u"sequence"]:
                    name = "%s/%s" % (node.name, child_sort)
                    grp_node = self.get_group_node_from_xml(child_obj, name,
                                                            uid_prefix=node.uid)
                    break
            if grp_node is not None:
                min_occur, max_occur = self.get_min_max_occur_from_xml(
                    child_obj, grp_node)
                node.sub_content = XsdUsage(grp_node, min_occur, max_occur)
            self.add_attributes_to_node(node, xml_obj)
        return node

    def fill_node_from_simple_content(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Complex type with simple content node that must be filled with
            the information available in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.types.XsdComplexType`
        :param xml_obj:
            XML element defining a simple content in a complex type in the
            XML Schema (typically, ``xsd:simpleContent``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.types.XsdComplexType`
        """
        self.collect_documentation(node, xml_obj)
        if xml_obj.find(u"{%s}restriction" % XSD_NS) is not None:
            der_obj = xml_obj.find(u"{%s}restriction" % XSD_NS)
            der_meth = RESTRICTION
            self.collect_documentation(node, der_obj)
            for child_obj in der_obj:
                ns, child_name = split_qname(child_obj.tag)
                if ns != XSD_NS or child_name not in CONSTRAINT_MAPPING:
                    continue
                node.add_constraint(child_name, child_obj.get(u"value"))
        elif xml_obj.find(u"{%s}extension" % XSD_NS) is not None:
            der_obj = xml_obj.find(u"{%s}extension" % XSD_NS)
            der_meth = EXTENSION
            self.collect_documentation(node, der_obj)
        else:
            raise XsdError("Simple Content in %s should contain either a "
                           "<restriction> or an <extension> child in the "
                           "%s XML Schema" % (node, self._xsd_file))
        type_nodes = self.get_type_nodes_from_xml(
            der_obj, u"base",
            uid_prefix=node.uid+u"/simpleContent/%s/%s" % (node.name, der_meth.lower()))
        if len(type_nodes) == 0:
            raise XsdError("%s in Simple Content in %s should contain either a"
                           "reference or an inner definition of type in %s XML"
                           "Schema" % (der_meth, node, self._xsd_file))
        node._base_type = type_nodes[0]
        node._deriv_method = der_meth
        self.add_attributes_to_node(node, der_obj)
        return node

    def fill_node_from_complex_content(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Complex type with complex content node that must be filled with
            the information available in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.types.XsdComplexType`
        :param xml_obj:
            XML element defining a complex content in a complex type in the
            XML Schema (typically, ``xsd:complexContent``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.types.XsdComplexType`
        """
        self.collect_documentation(node, xml_obj)
        if xml_obj.get(u"mixed", u"false") == u"true":
            node.content_type = MIXED
        if xml_obj.find(u"{%s}restriction" % XSD_NS) is not None:
            der_obj = xml_obj.find(u"{%s}restriction" % XSD_NS)
            der_meth = RESTRICTION
            self.collect_documentation(node, der_obj)
        elif xml_obj.find(u"{%s}extension" % XSD_NS) is not None:
            der_obj = xml_obj.find(u"{%s}extension" % XSD_NS)
            der_meth = EXTENSION
            self.collect_documentation(node, der_obj)
        else:
            raise XsdError("Complex Content in %s should contain either a "
                           "<restriction> or an <extension> child in the "
                           "%s XML Schema" % (node, self._xsd_file))
        type_nodes = self.get_type_nodes_from_xml(
            der_obj, u"base",
            uid_prefix=node.uid+u"/complexContent/%s/%s" % (node.name, der_meth.lower()))
        if len(type_nodes) == 0:
            raise XsdError("%s in Complex Content in %s should contain either "
                           "a reference or an inner definition of type in %s "
                           "XML Schema" % (der_meth, node, self._xsd_file))
        node._base_type = type_nodes[0]
        node._deriv_method = der_meth
        grp_node = None
        for child_obj in der_obj:
            ns, child_sort = split_qname(child_obj.tag)
            if ns == XSD_NS and child_sort == u"group":
                grp_node = self.get_group_node_from_xml(child_obj,
                                                        uid_prefix=node.uid)
                break
            if ns == XSD_NS and child_sort in [u"all", u"choice", u"sequence"]:
                name = "%s/complexContent/%s" % (node.name, child_sort)
                grp_node = self.get_group_node_from_xml(child_obj, name,
                                                        uid_prefix=node.uid)
                break
        if grp_node is not None:
            min_occur, max_occur = self.get_min_max_occur_from_xml(child_obj,
                                                                   grp_node)
            node.sub_content = XsdUsage(grp_node, min_occur, max_occur)
        self.add_attributes_to_node(node, der_obj)
        return node

    def fill_group_node(self, node, xml_obj):
        """
        Fills the ``node`` object with the information found in the ``xml_obj``
        XML element read in an XML Schema.

        :param node:
            Group node that must be filled with the information
            available in ``xml_obj``.
        :param node:
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        :param xml_obj:
            XML element defining a group in the XML Schema (typically,
            ``xsd:group``, ``xsd:all``, ``xsd:sequence``, ``xsd:choice``).
        :type xml_obj:
            :class:`~lxml.etree.Element`
        :returns:
            ``node`` after it has been filled.
        :rtype:
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        """
        self.collect_documentation(node, xml_obj)
        if xml_obj.tag == "{%s}group" % XSD_NS:
            # Finds the all, choice or sequence inside the group and uses it
            for child_obj in xml_obj:
                ns, child_sort = split_qname(child_obj.tag)
                if ns == XSD_NS and child_sort in [u"choice", u"all",
                                                   u"sequence"]:
                    xml_obj = child_obj
                    break
            else:
                return node
        grp_num = {u"all": 0, u"choice": 0, u"sequence": 0}
        for child_obj in xml_obj:
            ns, child_sort = split_qname(child_obj.tag)
            if ns == XSD_NS and child_sort == u"group":
                item_node = self.get_group_node_from_xml(child_obj,
                                                         uid_prefix=node.uid)
            elif ns == XSD_NS and child_sort in [u"all", u"choice",
                                                 u"sequence"]:
                name = u"%s/%s%d" % (node.name, child_sort, grp_num[child_sort])
                grp_num[child_sort] += 1
                item_node = self.get_group_node_from_xml(child_obj, name,
                                                         uid_prefix=node.uid)
            elif ns == XSD_NS and child_sort == u"element":
                item_node = self.get_element_node_from_xml(child_obj,
                                                           uid_prefix=node.uid)
            else:
                continue
            min_occur, max_occur = self.get_min_max_occur_from_xml(child_obj,
                                                                   item_node)
            node.add_item(item_node, min_occur, max_occur)
        return node

    def collect_documentation(self, node, xml_obj):
        """
        Adds documentation for the annotation XML elements into the node.
        """
        for xml_doc in xml_obj.xpath(u"xsd:annotation/xsd:documentation",
                                     namespaces={u"xsd": XSD_NS}):
            text = etree.tounicode(xml_doc)
            start = text.find(u">") + 1
            end = text.rfind(u"</")
            node.desc.append(text[start:end])



# -*- coding: utf-8 -*-

from collections import Sequence
import operator
from .utils import (ABSENT, DIFFERENT, SAME,
                              TEXT_ONLY, MIXED, ANY, ELEMENTS_ONLY, UNKNOWN)
from .graph_nodes import XMLElement, Group


class GraphComparator(object):
    def __init__(self, graph1, graph2):
        self.title1 = graph1.name
        self.title2 = graph2.name
        self.graph1 = graph1
        self.graph2 = graph2
        self.elt_cmprs = []
        self.cmpt_index = {}

    def compare(self):
        self.elt_cmprs = []
        self.cmpr_index = {}
        self.pair_elements()
        self.sort()
        for elt_cmpr in self.elt_cmprs:
            elt_cmpr.compare()

    def pair_elements(self):
        avail_elts1 = set(self.graph1.elements)
        avail_elts2 = set(self.graph2.elements)
        for elt1 in list(avail_elts1):
            poss_elts2 = self.graph2.find_all_elts(elt1.qname)
            if len(poss_elts2) == 0:
                self.add_elt_comparison(elt1, None)
            elif len(poss_elts2) == 1:
                self.add_elt_comparison(elt1, poss_elts2[0])
                avail_elts2.remove(poss_elts2[0])
            else:
                # There are several elements in graph2 with the same qname
                # as elt1. Chooses the one that has the most parents in common
                # with elt1 (better choice methods could be imagined).
                parents1 = set([elt.qname for elt in elt1.parent_elts])
                scores = []
                for elt2 in poss_elts2:
                    parents2 = set([elt.qname for elt in elt2.parent_elts])
                    comm_parents = parents1.intersection(parents2)
                    score = ( float(len(comm_parents)) / len(parents1)
                              + float(len(comm_parents)) / len(parents2) )
                    scores.append( (score, elt2) )
                scores.sort(key=operator.itemgetter(0), reverse=True)
                chosen_elt2 = scores[0][1]
                self.add_elt_comparison(elt1, chosen_elt2)
                avail_elts2.remove(chosen_elt2)
            avail_elts1.remove(elt1)
        for elt2 in list(avail_elts2):
            self.add_elt_comparison(None, elt2)
            avail_elts2.remove(elt2)
        assert len(avail_elts1) + len(avail_elts2) == 0

    def add_elt_comparison(self, elt1, elt2):
        assert elt1 is None or elt1 not in self.cmpr_index, \
            u"First element {0} already in a comparison".format(elt1)
        assert elt2 is None or elt2 not in self.cmpr_index, \
            u"Second element {0} already in a comparison".format(elt2)
        elt_cmpr = ElementComparator(self, elt1, elt2)
        self.elt_cmprs.append(elt_cmpr)
        if elt1 is not None:
            self.cmpr_index[elt1] = elt_cmpr
        if elt2 is not None:
            self.cmpr_index[elt2] = elt_cmpr

    def find_elt_cmpr(self, elt):
        return self.cmpr_index[elt]

    def sort(self):
        self.elt_cmprs.sort(key=operator.attrgetter(u"qname"))

    def __len__(self):
        return len(self.elts_comparisons)

    def __iter__(self):
        return iter(self.elts_comparisons)


class ElementComparator(object):
    def __init__(self, graph_cmpr, elt1, elt2):
        assert elt1 is not None or elt2 is not None
        self.graph_cmpr = graph_cmpr
        self.elt1 = elt1
        self.elt2 = elt2
        # Attributes
        self.attrs1_equality = {} # {XMLAttribute: CompResult}
        self.attrs2_equality = {} # {XMLAttribute: CompResult}
        # Content type
        self.ctype1_equality = None # CompResult
        self.ctype2_equality = None # CompResult
        self.text1_equality = None # CompResult
        self.text2_equality = None # CompResult
        # Sub-content
        self.subcnt1_equality = None # CompResult
        self.subcnt2_equality = None # CompResult
        # Parents
        self.parents1_equality = {} # {XMLElement: CompResult}
        self.parents2_equality = {} # {XMLElement: CompResult}

    @property
    def qname(self):
        if self.elt1 is not None:
            return self.elt1.qname
        else:
            return self.elt2.qname

    def compare(self):
        self.compare_attributes()
        self.compare_content_type()
        self.compare_textual_content()
        self.compare_sub_content()
        self.compare_parents()

    def compare_attributes(self):
        if self.elt1 is None:
            for att_occ2 in self.elt2.attributes.values():
                self.attrs2_equality[att_occ2.target] = CompResult(ABSENT)
        elif self.elt2 is None:
            for att_occ1 in self.elt1.attributes.values():
                self.attrs1_equality[att_occ1.target] = CompResult(ABSENT)
        else:
            for att_occ1 in self.elt1.attributes.values():
                att_occ2 = self.elt2.attributes.get(att_occ1.qname)
                if att_occ2 is None:
                    self.attrs1_equality[att_occ1.target] = CompResult(ABSENT)
                elif att_occ1 == att_occ2:
                    self.attrs1_equality[att_occ1.target] = CompResult(SAME)
                    self.attrs2_equality[att_occ2.target] = CompResult(SAME)
                else:
                    self.attrs1_equality[att_occ1.target] = \
                                                        CompResult(DIFFERENT)
                    self.attrs2_equality[att_occ2.target] = \
                                                        CompResult(DIFFERENT)
            for att_occ2 in self.elt2.attributes.values():
                if att_occ2.target in self.attrs2_equality:
                    continue
                self.attrs2_equality[att_occ2.target] = CompResult(ABSENT)

    def compare_content_type(self):
        if self.elt1 is None:
            self.ctype2_equality = CompResult(ABSENT)
        elif self.elt2 is None:
            self.ctype1_equality = CompResult(ABSENT)
        else:
            if ( self.elt1.content_type == self.elt2.content_type and
                 self.elt1.content_type != UNKNOWN):
                self.ctype1_equality = CompResult(SAME)
                self.ctype2_equality = CompResult(SAME)
            elif (self.elt2.content_type
                  in COMPATIBLE_CTYPES.get(self.elt1.content_type, [])):
                self.ctype1_equality = CompResult(DIFFERENT)
                self.ctype2_equality = CompResult(DIFFERENT)
            elif (self.elt1.content_type
                  in COMPATIBLE_CTYPES.get(self.elt2.content_type, [])):
                self.ctype1_equality = CompResult(DIFFERENT)
                self.ctype2_equality = CompResult(DIFFERENT)
            else:
                self.ctype1_equality = CompResult(ABSENT)
                self.ctype2_equality = CompResult(ABSENT)

    def compare_textual_content(self):
        if self.elt1 is None:
            self.text2_equality = CompResult(ABSENT)
        elif self.elt2 is None:
            self.text1_equality = CompResult(ABSENT)
        else:
            if ( self.elt1.content_type in (TEXT_ONLY, MIXED, ANY)
                 and self.elt2.content_type in (TEXT_ONLY, MIXED, ANY)):
                if ( (self.elt1.textual_content_type
                      == self.elt2.textual_content_type) and
                     (self.elt1.textual_content_is_list
                      == self.elt2.textual_content_is_list) and
                     (self.elt1.textual_content_values
                      == self.elt2.textual_content_values) ):
                    self.text1_equality = CompResult(SAME)
                    self.text2_equality = CompResult(SAME)
                else:
                    self.text1_equality = CompResult(DIFFERENT)
                    self.text2_equality = CompResult(DIFFERENT)
            else:
                self.text1_equality = CompResult(ABSENT)
                self.text2_equality = CompResult(ABSENT)

    def compare_sub_content(self):
        if self.elt1 is not None and self.elt1.content is not None:
            self.subcnt1_equality = \
                            build_results_struct(self.elt1.content.target)
        else:
            self.subcnt1_equality = CompResult(ABSENT)
        if self.elt2 is not None and self.elt2.content is not None:
            self.subcnt2_equality = \
                            build_results_struct(self.elt2.content.target)
        else:
            self.subcnt2_equality = CompResult(ABSENT)
        if self.elt1 is None or self.elt2 is None:
            return
        if ( not self.elt1.content_type in (MIXED, ELEMENTS_ONLY)
             or not self.elt2.content_type in (MIXED, ELEMENTS_ONLY) ):
            return
        self._compare_groups(self.elt1.content, self.elt2.content,
                             self.subcnt1_equality, self.subcnt2_equality)

    def _compare_groups(self, occur1, occur2, result1, result2,
                        differents=False):
        node1 = occur1.target
        node2 = occur2.target
        if occur1 == occur2:
            if not differents:
                result1.result = SAME
                result2.result = SAME
            else:
                result1.result = DIFFERENT
                result2.result = DIFFERENT
        elif node1.__class__ == node2.__class__:
            result1.result = DIFFERENT
            result2.result = DIFFERENT
        elif len(node1) == 1 or len(node2) == 1:
            if ( occur1.minimum == occur2.minimum
                 and occur1.maximum == occur2.maximum
                 and not differents):
                result1.result = SAME
                result2.result = SAME
            else:
                result1.result = DIFFERENT
                result2.result = DIFFERENT
        else:
            differents = True
        not_found1 = dict([(sub_occur, result1.sub_results[idx])
                           for (idx, sub_occur) in enumerate(node1)])
        not_found2 = dict([(sub_occur, result2.sub_results[idx])
                           for (idx, sub_occur) in enumerate(node2)])
        sub_not_found1 = {}
        sub_not_found2 = {}
        idx2 = -1
        for sub_occur1 in node1:
            sub_node1 = sub_occur1.target
            sub_node2_match, sub_nodes_compare = \
                                self._build_match_compare_functions(sub_node1)
            for tmp_idx in range(idx2+1, len(node2)):
                if sub_node2_match(node2[tmp_idx].target):
                    idx2 = tmp_idx
                    sub_result1 = not_found1.pop(sub_occur1)
                    sub_result2 = not_found2.pop(node2[tmp_idx])
                    nfnd1, nfnd2 = sub_nodes_compare(
                        sub_occur1, node2[tmp_idx],
                        sub_result1, sub_result2, differents)
                    sub_not_found1.update(nfnd1)
                    sub_not_found2.update(nfnd2)
                    break
            else:
                # Search corresponding element before current index (idx2)
                for tmp_idx in range(0, idx2):
                    if ( sub_node2_match(node2[tmp_idx].target)
                         and node2[tmp_idx] in not_found2):
                        sub_result1 = not_found1.pop(sub_occur1)
                        sub_result2 = not_found2.pop(node2[tmp_idx])
                        nfnd1, nfnd2 = sub_nodes_compare(
                            sub_occur1, node2[tmp_idx],
                            sub_result1, sub_result2, differents)
                        if isinstance(node1, Sequence):
                            # Order matters and we found elt2 before current
                            # index (idx2) so the result can't be "SAME"
                            sub_result1.result = DIFFERENT
                            sub_result2.result = DIFFERENT
                        sub_not_found1.update(nfnd1)
                        sub_not_found2.update(nfnd2)
                        break
        # Collect the elements not found in the sub-groups that didn't match
        # anything (sub-elt1 and sub-elt2)
        for sub_occur1, sub_result1 in not_found1.items():
            if isinstance(sub_occur1.target, Group):
                sub_not_found1.update(
                    self._collect_all_not_found(sub_occur1, sub_result1))
        for sub_occur2, sub_result2 in not_found2.items():
            if isinstance(sub_occur2.target, Group):
                sub_not_found2.update(
                    self._collect_all_not_found(sub_occur2, sub_result2))
        # Try to match the not found elements (elt1) with the elements not
        # found in the sub-groups (sub-elt2)
        for sub_occur1 in not_found1.keys():
            sub_node1 = sub_occur1.target
            if isinstance(sub_node1, Group):
                continue
            sub_node2_match, sub_nodes_compare = \
                                self._build_match_compare_functions(sub_node1)
            for sub_occur2 in sub_not_found2.keys():
                if sub_node2_match(sub_occur2.target):
                    sub_result1 = not_found1.pop(sub_occur1)
                    sub_result2 = sub_not_found2.pop(sub_occur2)
                    _, _ = sub_nodes_compare(sub_occur1, sub_occur2,
                                             sub_result1, sub_result2, True)
                    break
        # Try to match the elements not found in the sub-groups (sub-elt1) with
        # the not found elements (elt2)
        for sub_occur1 in sub_not_found1.keys():
            sub_node1 = sub_occur1.target
            if isinstance(sub_node1, Group):
                continue
            sub_node2_match, sub_nodes_compare = \
                                self._build_match_compare_functions(sub_node1)
            for sub_occur2 in not_found2.keys():
                if sub_node2_match(sub_occur2.target):
                    sub_result1 = sub_not_found1.pop(sub_occur1)
                    sub_result2 = not_found2.pop(sub_occur2)
                    _, _ = sub_nodes_compare(sub_occur1, sub_occur2,
                                             sub_result1, sub_result2, True)
                    break
        # Add the elements not found in the sub-groups to the elements not
        # found in this group; return them
        not_found1.update(sub_not_found1)
        not_found2.update(sub_not_found2)
        return not_found1, not_found2

    def _compare_elts(self, occur1, occur2, result1, result2, differents=False):
        if occur1 == occur2 and not differents:
            result1.result = SAME
            result2.result = SAME
        else:
            result1.result = DIFFERENT
            result2.result = DIFFERENT
        return [], []

    def _build_match_compare_functions(self, sub_node1, groups=True):
        if isinstance(sub_node1, XMLElement):
            assoc_elt2 = self.graph_cmpr.find_elt_cmpr(sub_node1).elt2
            def sub_node2_match(sub_node2):
                return  sub_node2 == assoc_elt2
            sub_nodes_compare = self._compare_elts
        elif groups and isinstance(sub_node1, Group):
            class2 = sub_node1.__class__
            def sub_node2_match(sub_node2):
                return isinstance(sub_node2, class2)
            sub_nodes_compare = self._compare_groups
        else:
            def sub_node2_match(subnode_idx):
                return False
            def sub_nodes_compare(s, o1, o2, r1, r2):
                return None
        return sub_node2_match, sub_nodes_compare

    def _collect_all_not_found(self, occur, result):
        gr_node = occur.target
        if not isinstance(gr_node, Group):
            return {}
        not_found = {}
        for idx, sub_occur in enumerate(gr_node):
            not_found[sub_occur] = result.sub_results[idx]
            if isinstance(sub_occur.target, Group):
                not_found.update(
                    self._collect_all_not_found(sub_occur,
                                                result.sub_results[idx]))
        return not_found

    def compare_parents(self):
        if self.elt1 is None:
            for parent2 in self.elt2.parent_elts:
                self.parents2_equality[parent2] = CompResult(ABSENT)
        elif self.elt2 is None:
            for parent1 in self.elt1.parent_elts:
                self.parents1_equality[parent1] = CompResult(ABSENT)
        else:
            for parent1 in self.elt1.parent_elts:
                poss_par2 = self.elt2.find_parents(parent1.qname)
                if len(poss_par2) == 0:
                    self.parents1_equality[parent1] = CompResult(ABSENT)
                else:
                    assoc_par2 = self.graph_cmpr.find_elt_cmpr(parent1).elt2
                    if assoc_par2 in poss_par2:
                        self.parents1_equality[parent1] = CompResult(SAME)
                        self.parents2_equality[assoc_par2] = CompResult(SAME)
                    else:
                        self.parents1_equality[parent1] = CompResult(ABSENT)
            for parent2 in self.elt2.parent_elts:
                if parent2 in self.parents2_equality:
                    continue
                self.parents2_equality[parent2] = CompResult(ABSENT)


COMPATIBLE_CTYPES = {
    TEXT_ONLY: (MIXED, ANY),
    ELEMENTS_ONLY: (MIXED, ANY),
    MIXED: (TEXT_ONLY, ELEMENTS_ONLY, ANY),
    ANY: (TEXT_ONLY, ELEMENTS_ONLY, MIXED)
}


class CompResult(object):
    def __init__(self, result, is_group=False):
        assert result in (SAME, ABSENT, DIFFERENT)
        self.is_group = is_group
        self.result = result
        self.sub_results = []

    def __eq__(self, other):
        return self.result == other

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return u"<Comparison: {0} {1}>".format(self.result, self.sub_results)

def build_results_struct(group):
    result = CompResult(ABSENT, True)
    for sub_occur in group:
        if isinstance(sub_occur.target, XMLElement):
            result.sub_results.append(CompResult(ABSENT))
        elif isinstance(sub_occur.target, Group):
            result.sub_results.append(build_results_struct(sub_occur.target))
    return result





"""
Module defining a specific exception for the errors occuring while analyzing
a DTD.

.. autoclass:: DtdError
   :show-inheritance:
"""

class DtdError(Exception):
    """
    Exception raised when finding an error while analyzing a DTD.
    """
    pass




# -*- coding: utf-8 -*-
"""
Sub-package containing the code that can read a DTD and return
a :class:`~dtd.root.DtdDefinition` object that
contains the nodes (elements, attributes, entities, etc.) extracted from this
DTD.

External programs should just call the
:func:`~dtd.read_dtd` method defined
at the top level of this package.

The other modules of this package contain all the classes and
functions necessary to read the DTD.

.. autofunction:: read_dtd
"""
__docformat__ = "restructuredtext en"


from .parser import DtdParser


def read_dtd(dtd_file, ns_mapping=None, qualify_local_elements=False,
              qualify_local_attributes=False):
    """
    Reads the ``xsd_dtd`` and returns a DtdDefinition object containing
    the nodes read in the DTD.

    The DTD doesn't support namespaces. The elements and attributes have to
    be declared with their prefix (e.g. ``xlink:href``). The ``ns_mapping``
    parameter allows the building of qualified names (namespace plus local
    name) from the DTD declatation.

    :param dtd_file:
        Name of the file that contains the DTD.
    :type dtd_file:
        :class:`unicode`
    :param ns_mapping:
        Mapping of prefixes towards namespaces ``{prefix: namespace}``. It is
        possible to give a default namespace by associating a namespace with
        ``u""`` or ``None``.
    :type ns_mapping:
        dict {:class:`unicode`: :class:`unicode`}
    :param qualify_local_elements:
        Flag indicating that the elements without a prefix must be qualified
        with the default namespace given in ``ns_mapping`` (namespace
        indexed with ``u""``).
    :type qualify_local_elements:
        :class:`bool`
    :param qualify_local_attributes:
        Flag indicating that the attributes without a prefix must be qualified
        with the default namespace given in ``ns_mapping`` (namespace
        indexed with ``u""``).
    :type qualify_local_attributes:
        :class:`bool`
    :returns:
        A DtdDefinition object grouping all the nodes read in the DTD.
    :rtype:
        :class:`~dtd.root.DtdDefinition`
    """
    prs = DtdParser(ns_mapping)
    prs.parse_dtd(dtd_file, qualify_local_elements, qualify_local_attributes)
    return prs.dtd_def



# -*- coding: utf-8 -*-
"""
Sub-package containing the code that can read an XML Schema and return
an :class:`~pyxst.xml_struct.xsd.root.XsdSchema` object that
contains the nodes (elements, attributes, types, etc.) extracted from this
XML Schema.

External programs should just call the
:func:`~pyxst.xml_struct.xsd.read_schema` method defined
at the top level of this package.

The other modules of this package contain all the classes and
functions necessary to read the XML Schema:

:mod:`~pyxst.xml_struct.xsd.parser`
    Definition of the function that can parse an XML Schema and return a 
    ``XsdSchema`` object
:mod:`~pyxst.xml_struct.xsd.root`
    Definition of the ``XsdSchema`` object that contains all the information 
    read in an XML Schema and numerous functions to extract this information
:mod:`~pyxst.xml_struct.xsd.elements`
    Definition of the element node that represents an element definition read
    in an XML Schema
:mod:`~pyxst.xml_struct.xsd.attributes`
    Definition of the attribute node and the attribute group node that 
    represent the corresponding definitions read in an XML Schema
:mod:`~pyxst.xml_struct.xsd.types`
    Definition of the type nodes that represent the definition of types read
    in an XML Schema
:mod:`~pyxst.xml_struct.xsd.constraints`
    Definition of the constraints that can be applied to the type nodes and 
    that represent the corresponding constraints read in an XML Schema
:mod:`~pyxst.xml_struct.xsd.groups`
    Definition of the group nodes that represent the definition of groups read
    in an XML Schema (a group is set of sub-elements that can occur in a type,
    typically a sequence, a choice or an all).
:mod:`~pyxst.xml_struct.xsd.utils`
    Definition of the several useful classes and functions.
:mod:`~pyxst.xml_struct.xsd.errors`
    Definition of a dedicated exception.

.. autofunction:: read_schema
"""
__docformat__ = "restructuredtext en"


from .parser import parse_schema


def read_schema(xsd_file):
    """
    Reads the ``xsd_file`` and returns an XsdSchema object containing
    the nodes (cf. nodes derived from
    :class:`~pyxst.xml_struct.xsd.utils.XsdNode`) read in
    the XML Schema.

    These nodes represent all the information read inside the XSD Schema
    (definition of XML elements and XML attributes, declaration of types,
    groups and attribute groups).

    :param xsd_file:
        Name of the file that contains the XML Schema.
    :type xsd_file:
        :class:`str`
    :returns:
        An XsdSchema object grouping all the nodes read from the XML Schema.
    :rtype:
        :class:`~pyxst.xml_struct.xsd.root.XsdSchema`
    """
    return parse_schema(xsd_file)



# -*- coding: utf-8 -*-
"""
Module containing the parser function that can read a DTD file and
return a :class:`~dtd.root.DtdDefinition` object.

.. autofunction:: parse_dtd
"""

from traceback import format_exc
from os import path as osp

from lxml import etree
from six import text_type

from .errors import DtdError
from .root import DtdDefinition
from .nodes import DtdElement, DtdAttribute
from .groups import DtdSequence, DtdChoice
from .utils import resolve_xml_qname, DtdUsage
from . import constants


IMPLIED = u"IMPLIED"
FIXED = u"FIXED"
EMPTY = u"EMPTY"
ANY = u"ANY"
MIXED = u"MIXED"
ELEMENT = u"ELEMENT"
PCDATA = u"PCDATA"
OPTIONAL = u"OPT"
MULTIPLE = u"MULT"
PLUS = u"PLUS"
CHOICE = u"OR"
SEQUENCE = u"SEQ"

CONTENT_TYPE_MAPPING = {
    EMPTY : constants.EMPTY,
    ANY: constants.ANY,
    MIXED: constants.MIXED,
    ELEMENT: constants.ELEMENTS_ONLY
}

GROUP_MAPPING = {
    CHOICE: DtdChoice,
    SEQUENCE: DtdSequence,
}


class DtdParser(object):
    def __init__(self, ns_mapping=None):
        if ns_mapping is None:
            self.ns_mapping = {}
        else:
            self.ns_mapping = ns_mapping
        self.qualif_local_elt = True
        self.qualif_local_att = False
        self.dtd_def = None

    def parse_dtd(self, dtd_file, qualify_local_elements=False,
                  qualify_local_attributes=False):
        self.qualif_local_elt = qualify_local_elements
        self.qualif_local_att = qualify_local_attributes
        # Opens the XML Schema file
        try:
            dtd_file = osp.abspath(dtd_file)
        except Exception:
            pass
        try:
            dtd_decl = etree.DTD(dtd_file)
        except etree.DTDParseError:
            raise DtdError("Found a syntax error while reading the %s DTD "
                           "file\n%s" % (dtd_file, format_exc()))
        self.dtd_def = DtdDefinition()
        for elt_decl in dtd_decl.iterelements():
            elt_obj = self.build_element(elt_decl)
            self.dtd_def.add_element(elt_obj)
        for elt_decl in dtd_decl.iterelements():
            self.fill_element(elt_decl)

    def build_element(self, elt_decl, uid_prefix=u""):
        elt_name = self.build_name(elt_decl, self.qualif_local_elt)
        elt_obj = DtdElement(elt_name, uid_prefix=uid_prefix)
        # Attributes
        for att_decl in elt_decl.iterattributes():
            att_obj = self.build_attribute(att_decl, uid_prefix=elt_obj.uid)
            min_occur = 1
            if att_decl.default.upper() == IMPLIED:
                min_occur = 0
            elt_obj.add_attribute(att_obj, min_occur, 1)
        return elt_obj

    def build_attribute(self, att_decl, uid_prefix):
        att_name = self.build_name(att_decl, self.qualif_local_att)
        att_obj = DtdAttribute(att_name, uid_prefix=uid_prefix)
        att_obj.dtd_type = att_decl.type.upper()
        att_obj.fixed_value = (att_decl.default.upper() == FIXED)
        if att_decl.default_value is not None:
            att_obj.default_value = att_decl.default_value
        poss_vals = [text_type(val) for val in att_decl.values()]
        if len(poss_vals) > 0:
            att_obj.dtd_type = u"CDATA"
            att_obj.possible_values = poss_vals
        return att_obj

    def fill_element(self, elt_decl, uid_prefix=u""):
        elt_name = self.build_name(elt_decl, self.qualif_local_elt)
        elt_obj = self.dtd_def.elements.get(
            u"{0}/{1}".format(uid_prefix, elt_name))
        if elt_obj is None:
            raise DtdError(u"{0} should have been already built"
                           u"".format(elt_name))
        # Sub-content
        elt_obj.content_type = CONTENT_TYPE_MAPPING.get(elt_decl.type.upper())
        if elt_decl.content is not None:
            elt_obj.sub_content = self.build_group(elt_decl.content,
                                                   uid_prefix=elt_obj.uid)
        if ( elt_obj.content_type == constants.MIXED
             and elt_obj.sub_content is None):
            elt_obj.content_type = constants.TEXT_ONLY

    def build_group(self, content_decl, uid_prefix=u"", parent_grp=None,
                    sub_grp_counts=None):
        if content_decl.type.upper() == PCDATA:
            return
        min_occur = 1
        if content_decl.occur.upper() in (OPTIONAL, MULTIPLE):
            min_occur = 0
        max_occur = 1
        if content_decl.occur.upper() in (MULTIPLE, PLUS):
            max_occur = constants.INFINITY
        if content_decl.type.upper() == ELEMENT:
            ns_map = self.ns_mapping.copy()
            if not(self.qualif_local_elt):
                if None in ns_map:
                    ns_map.pop(None)
                if u"" in ns_map:
                    ns_map.pop(u"")
            elt_name = resolve_xml_qname(content_decl.name, ns_map)
            elt_obj = self.dtd_def.elements.get(u"/{0}".format(elt_name))
            if elt_obj is None:
                raise DtdError(
                    u"In {0}, the content refers to {1} element that can't be "
                    u"found".format(uid_prefix, elt_name))
            if parent_grp is None:
                # Happens when there is only one child element inside the
                # element
                grp = DtdSequence(u"sequence", uid_prefix=uid_prefix)
                grp.add_item(elt_obj, min_occur, max_occur)
                return DtdUsage(grp, 1, 1)
            else:
                parent_grp.add_item(elt_obj, min_occur, max_occur)
        elif content_decl.type.upper() in (CHOICE, SEQUENCE):
            cnt_type = content_decl.type.upper()
            cnt_class = GROUP_MAPPING[cnt_type]
            if parent_grp is None:
                grp = cnt_class(u"{0}0".format(cnt_type.lower()),
                                uid_prefix=uid_prefix)
                cnts = {CHOICE: 0, SEQUENCE: 0}
                self.build_group(content_decl.left, uid_prefix=uid_prefix,
                                 parent_grp=grp, sub_grp_counts=cnts)
                self.build_group(content_decl.right, uid_prefix=uid_prefix,
                                 parent_grp=grp, sub_grp_counts=cnts)
                return DtdUsage(grp, min_occur, max_occur)
            elif (isinstance(parent_grp, cnt_class) and min_occur == 1
                  and max_occur == 1):
                # Adds the group content directly inside the parent group
                # cf. tree structure of the content in lxml.DTD
                self.build_group(
                    content_decl.left, uid_prefix=uid_prefix,
                    parent_grp=parent_grp, sub_grp_counts=sub_grp_counts)
                self.build_group(
                    content_decl.right, uid_prefix=uid_prefix,
                    parent_grp=parent_grp, sub_grp_counts=sub_grp_counts)
            else:
                # Builds a new group and inserts it in the parent group
                grp = cnt_class(
                    u"{0}{1}".format(cnt_type.lower(),sub_grp_counts[cnt_type]),
                                uid_prefix=parent_grp.uid)
                sub_grp_counts[cnt_type] += 1
                cnts = {CHOICE: 0, SEQUENCE: 0}
                self.build_group(
                    content_decl.left, uid_prefix=uid_prefix,
                    parent_grp=grp, sub_grp_counts=cnts)
                self.build_group(
                    content_decl.right, uid_prefix=uid_prefix,
                    parent_grp=grp, sub_grp_counts=cnts)
                parent_grp.add_item(grp, min_occur, max_occur)
        else:
            raise DtdError(u"Unexpected type of content in {0}: {1}"
                           u"".format(uid_prefix, content_decl.type))

    def build_name(self, decl, qualif_item):
        """
        Builds the name of an element of an attribute using the namespace
        mapping
        """
        if decl.prefix != u"" and decl.prefix is not None:
            namespace = self.ns_mapping.get(decl.prefix, u"")
        elif qualif_item:
            namespace = self.ns_mapping.get(u"", self.ns_mapping.get(None, u""))
        else:
            namespace = u""
        if namespace == u"":
            return text_type(decl.name)
        else:
            return u"{{{0}}}{1}".format(namespace, decl.name)



# -*- coding: utf-8 -*-
"""
This module contains the class that defines the structure of the XML documents
that will be returned by the :mod:`xml_struct` module.

This structure is a graph that contains different types of nodes,
mainly the XML elements and the XML attributes defined in
:mod:`xml_struct.graph_nodes` module.

.. autoclass:: XMLGraph
   :show-inheritance:

.. autofunction:: compute_textual_content

.. autodata:: TYPES_MAPPING
"""
__docformat__ = "restructuredtext en"


import operator

from ..xsd import types as xsdTypes
from ..xsd import utils as xsdUtils
from ..xsd.types import XsdSimpleListType, XsdSimpleUnionType, XsdComplexType
from ..xsd.groups import XsdGroup, XsdSequence, XsdAll
from ..dtd import constants as dtdConstants
from ..dtd.groups import DtdSequence, DtdGroup

from .utils import (
    MIXED, INFINITY, UNKNOWN, BOOLEAN, INTEGER, FLOAT, STRING, DATETIME,
    DURATION, BASE64, HEXA, ANY)
from .graph_nodes import XMLElement, XMLAttribute, Group, Sequence, Alternative
from .graph_target_relations import RELATION_CLASS


class XMLGraph(object):
    """
    Class that contains the graph representing the XML structure.

    The graph only contains
    :class:`~pyxst.xml_struct.graph.XMLElement` objects
    that contain themselves all the other nodes (groups of child
    elements, attributes).

    .. attribute:: name

       Name of the graph.

       Type: :class:`unicode`

    .. attribute:: elements

       Dictionary containing the elements defined in the graph. The elements
       are indexed with their identifier.

       Type: :class:`dict` of
       :class:`~pyxst.xml_struct.graph.XMLElement`
       indexed with :class:`unicode`

    .. attribute:: target_relations

       List of the target relations that are defined between the nodes
       of the graph.

       Target relations describe relations that can be translated into some
       arKItect concepts.

       Type: :class:`list` of
       :class:`~pyxst.xml_struct.graph_target_relations.TargetRelation`

    .. attribute:: _high_level_elts

       Private attribute that contains the XSD elements returned by the schema
       reader that are high level elements (elements defined at the top level
       of the schema).

       This set is only used while building the graph from the schema.

       Type: :class:`set` of
       :class:`~pyxst.xml_struct.xsd.elements.XsdElement`

    .. automethod:: __init__

    .. automethod:: __iter__

    .. automethod:: __len__
    """
    def __init__(self, name, source=u"", source_type=None, type_mapper=None):
        """
        Initializes a new graph.

        :param name:
            Name of the graph.
        :type name:
            :class:`unicode`
        """
        self.name = name
        self.source = source
        self.source_type = source_type
        if type_mapper is None:
            type_mapper = DefaultTypeMapper()
        self.type_mapper = type_mapper
        self.elements = []
        self.elts_index = {} # {qname: [XMLElement]}
        self.target_relations = []
        self._high_level_elts = set()
        # {qname: [(XMLElement, XsdElement.uid or DtdElement.uid, XsdType.uid or None)]}
        self._elts_def_index = {}
        # These attributes are used to keep track of the parameters used to
        # build the graph from a DTD (the DTD parser might have added namespaces
        # to the elements)
        self.dtd_ns_mapping = {}
        self.dtd_qualif_local_elt = False
        self.dtd_qualif_local_attr = False
        # The following attribute is used for analyzing a library of XML
        # documents whose structure is given by the graph. It is used to
        # count the number of times an element is actually used as a root
        # element in all the XML documents of the library.
        # (Some elements might never be used in the documents)
        self.root_uses = {} # {XMLElement: {xml_filename: uses_count},]}

    def __iter__(self):
        """
        Returns an iterator over the elements contained in the graph.

        :returns:
            An iterator over all the elements defined in the graph.
        :rtype:
            :class:`listiterator`
        """
        return iter(self.elements)

    def __len__(self):
        """
        Returns the number of elements in the graph.

        :returns:
            The number of elements defined in the graph.
        :rtype:
            :class:`int`
        """
        return len(self.elements)

    def find_all_elts(self, qname):
        return self.elts_index.get(qname, [])

    def find_high_level_elts(self, qname):
        result = []
        for elt in self.elts_index.get(qname, []):
            if elt.is_high_level:
                result.append(elt)
        return result

    def add_target_relation(self, relation_name, start_node, end_node):
        """
        Adds a target relation between ``start_node`` and ``end_node``.

        The target relation describes some relations between the graph
        nodes that can be translated into some arKItect concepts.

        :param relation_name:
            Name of the relation. This name will be used to get the class
            of the target relation thanks to the ``RELATION_CLASS``
            dictionary.
        :type relation_name:
            :class:`str`
        :param start_node:
            XML node where the target relation starts from.
        :type node:
            :class:`~pyxst.xml_struct.graph_nodes.XMLNode`
        :param end_node:
            XML node where the target relation ends to.
        :type node:
            :class:`~pyxst.xml_struct.graph_nodes.XMLNode`
        """
        relation = RELATION_CLASS[relation_name](start_node, end_node)
        self.target_relations.append(relation)

    def build_from_schema(self, schema):
        """
        Builds the graph from the ``schema`` object.

        This method will create all the elements, attributes and sub-elements
        that are defined in the XML Schema. It starts with the high-level
        elements declared at the top level of the schema and then goes on with
        the hierarchy of sub-elements in each element.

        :param schema:
            Schema object returned by the schema parser.
        :type schema:
            :class:~pyxst.xml_struct.xsd.root.XsdSchema`
        """
        self._high_level_elts = set(schema.elements.values())
        for xsd_elt in schema.elements.values():
            self.build_xsd_element(xsd_elt)
        self._high_level_elts = set()
        self.finalize()

    def build_xsd_element(self, xsd_elt, parent_elt=None):
        """
        Builds an XML element from the information defined in the ``xsd_elt``
        extracted from the schema object.

        This method creates a new element or returns a previously declared
        element (element already created in the graph).

        When creating a new element, this method creates the XML element itself,
        its attributes and the group that contains its sub-elements.

        :param xsd_elt:
            Element definition extracted from the XML Schema.
        :type xsd_elt:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement`
        :param parent_elt:
            Xml element of the graph that will be the parent of the newly
            created XML element.
        :type parent_elt:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        :returns:
            Xml element that corresponds to the ``xsd_elt`` definition in the
            XML Schema and that is created and filled inside this method.
        :rtype:
           :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        exist_elts = self._elts_def_index.get(xsd_elt.name, [])
        for gr_elt, xsd_elt_uid, xsd_type_uid in exist_elts:
            # In the list of graph elements with the same name, is there
            # an element created from an xsd_elt with the same uid or an
            # element created with an xsd_elt whose xsd_type has the same uid
            # (in both cases, it is the same element as the one required here)
            if xsd_elt.uid == xsd_elt_uid:
                if parent_elt is not None:
                    gr_elt.add_parent_elt(parent_elt)
                return gr_elt
            elif xsd_elt.xsd_type.uid == xsd_type_uid:
                if parent_elt is not None:
                    gr_elt.add_parent_elt(parent_elt)
                return gr_elt
        # Creates the new graph element
        gr_elt = XMLElement(xsd_elt.name,
                            (xsd_elt in self._high_level_elts),
                            desc= xsd_elt.desc,)
        if parent_elt is not None:
            gr_elt.add_parent_elt(parent_elt)
        # Adds the element in the list of elements and in the index
        self.elements.append(gr_elt)
        self._elts_def_index.setdefault(xsd_elt.name, [])
        self._elts_def_index[xsd_elt.name].append(
            (gr_elt, xsd_elt.uid, xsd_elt.xsd_type.uid) )
        # Sets the content type of the graph element
        typ, lst, vals = compute_xsd_textual_content(xsd_elt.xsd_type, self.type_mapper)
        gr_elt.textual_content_type = typ
        gr_elt.textual_content_is_list = lst
        gr_elt.textual_content_values = vals
        if isinstance(xsd_elt.xsd_type, XsdComplexType):
            # Creates and adds the attributes on the graph element
            for xsd_usg in xsd_elt.xsd_type.attributes:
                if xsd_usg.max_occur == 0:
                    continue
                xsd_att = xsd_usg.referenced_node
                gr_att = XMLAttribute(xsd_att.name)
                typ, lst, vals = compute_xsd_textual_content(xsd_att.xsd_type, self.type_mapper)
                gr_att.textual_content_type = typ
                gr_att.textual_content_is_list = lst
                gr_att.textual_content_values = vals
                gr_elt.add_attribute(gr_att,
                                     xsd_usg.min_occur, xsd_usg.max_occur)
            # Processes the content (sub-elements)
            if xsd_elt.xsd_type.sub_content is not None:
                xsd_grp = xsd_elt.xsd_type.sub_content.referenced_node
                xsd_min = xsd_elt.xsd_type.sub_content.min_occur
                xsd_max = xsd_elt.xsd_type.sub_content.max_occur
                if xsd_max == xsdUtils.INFINITY:
                    xsd_max = INFINITY
                gr_grp = self.build_sub_elements_xsd_group(xsd_grp, xsd_max,
                                                       parent_elt=gr_elt)
                gr_elt.set_content(gr_grp, xsd_min, xsd_max)
        return gr_elt

    def build_sub_elements_xsd_group(self, xsd_group, max_occur, parent_elt=None):
        """
        Builds a group of the graph that contains XML elements that are
        children of an Xml element (``parent_elt``) and that is defined in
        the XML Schema with ``xsd_group``.

        This method creates a new group and adds in this group the sub-elements
        declared in ``xsd_group``. Several types of group can exist depending on
        the ``xsd_group`` (sequence, all, choice) and on the maximum occurence
        of the group.

        Inside a group, we can find sub-groups or sub-elements. This method
        recursively calls the methods of the graph to build the sub-elements
        and the sub-groups.

        :param xsd_group:
            Group definition extracted from the XML Schema.
        :type xsd_group:
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        :param max_occur:
            Maximum occurence of the ``xsd_group`` inside its parent. Can
            be ``INFINITY``.
        :type max_occur:
            :class:`int` or :data:`~pyxst.xml_struct.utils.INFINITY`
        :param parent_elt:
            Xml element of the graph that will be the parent of the newly
            created group and its children.
        :type parent_elt:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        :returns:
            Group that corresponds to the ``xsd_group`` definition in the
            XML Schema and that is created and filled inside this method.
        :rtype:
           :class:`~pyxst.xml_struct.graph_nodes.Group`
        """
        # Builds the group object that will be inserted in the graph
        if len(xsd_group) == 1:
            # If there is only one child in the group, the group is always a
            # simple group without order.
            gr_grp = Group(parent_elt)
        elif isinstance(xsd_group, XsdSequence):
            gr_grp = Sequence(parent_elt)
        elif isinstance(xsd_group, XsdAll):
            gr_grp = Group(parent_elt)
        else:
            gr_grp = Alternative(parent_elt)
        # Creates the child elements of the group and adds them in the group
        # object to be inserted in the graph
        for xsd_usg in iter(xsd_group):
            if xsd_usg.max_occur == 0:
                continue
            max_occur =  xsd_usg.max_occur
            if max_occur == xsdUtils.INFINITY:
                max_occur = INFINITY
            if isinstance(xsd_usg.referenced_node, XsdGroup):
                gr_child = self.build_sub_elements_xsd_group(
                    xsd_usg.referenced_node, max_occur,
                    parent_elt=parent_elt)
            else:
                gr_child = self.build_xsd_element(xsd_usg.referenced_node,
                                                  parent_elt=parent_elt)
            gr_grp.add_child(gr_child, xsd_usg.min_occur, max_occur)
        return gr_grp

    def build_from_dtd(self, dtd_def):
        """
        Builds the graph from the ``dtd_def`` object.

        This method will create all the elements, attributes and sub-elements
        that are defined in the DTD.

        :param dtd_def:
            DTD definition object returned by the DTD parser.
        :type schema:
            :class:~dtd.root.DtdDefinition`
        """
        for dtd_elt in dtd_def.elements.values():
            self.build_dtd_element(dtd_elt)
        self.finalize()

    def build_dtd_element(self, dtd_elt, parent_elt=None):
        """
        Builds an XML element from the information defined in the ``dtd_elt``
        extracted from the DTD Definition object.

        This method creates a new element or returns a previously declared
        element (element already created in the graph).

        When creating a new element, this method creates the XML element itself,
        its attributes and the group that contains its sub-elements.

        :param dtd_elt:
            Element definition extracted from the XML Schema.
        :type dtd_elt:
            :class:`~dtd.elements.DtdElement`
        :param parent_elt:
            Xml element of the graph that will be the parent of the newly
            created XML element.
        :type parent_elt:
            :class:`~xml_struct.graph_nodes.XMLElement`
        :returns:
            Xml element that corresponds to the ``dtd_elt`` definition in the
            DTD and that is created and filled inside this method.
        :rtype:
           :class:`~xml_struct.graph_nodes.XMLElement`
        """
        exist_elts = self._elts_def_index.get(dtd_elt.name, [])
        for gr_elt, dtd_elt_uid, _ in exist_elts:
            # In the list of graph elements with the same name, is there
            # an element created from a dtd_elt with the same uid
            # (then, it is the same element as the one required here)
            if dtd_elt.uid == dtd_elt_uid:
                if parent_elt is not None:
                    gr_elt.add_parent_elt(parent_elt)
                return gr_elt
        # Creates the new graph element
        gr_elt = XMLElement(dtd_elt.name,
                            True) # All elements are high-level elements
        if parent_elt is not None:
            gr_elt.add_parent_elt(parent_elt)
        # Adds the element in the list of elements and in the index
        self.elements.append(gr_elt)
        self._elts_def_index.setdefault(dtd_elt.name, [])
        self._elts_def_index[dtd_elt.name].append(
            (gr_elt, dtd_elt.uid, None) )
        # Sets the content type of the graph element
        typ, lst = compute_dtd_textual_content(dtd_elt.content_type)
        gr_elt.textual_content_type = typ
        gr_elt.textual_content_is_list = lst
        # Creates and adds the attributes on the graph element
        for dtd_usg in dtd_elt.attributes.values():
            if dtd_usg.max_occur == 0:
                continue
            dtd_att = dtd_usg.referenced_node
            gr_att = XMLAttribute(dtd_att.name)
            typ, lst = compute_dtd_textual_content(dtd_att.dtd_type)
            gr_att.textual_content_type = typ
            gr_att.textual_content_is_list = lst
            if dtd_att.possible_values is not None:
                gr_att.textual_content_values = set(dtd_att.possible_values)
            gr_elt.add_attribute(gr_att, dtd_usg.min_occur, dtd_usg.max_occur)
        # Processes the content (sub-elements)
        if dtd_elt.sub_content is not None:
            dtd_grp = dtd_elt.sub_content.referenced_node
            dtd_min = dtd_elt.sub_content.min_occur
            dtd_max = dtd_elt.sub_content.max_occur
            if dtd_max == dtdConstants.INFINITY:
                dtd_max = INFINITY
            gr_grp = self.build_sub_elements_dtd_group(dtd_grp, dtd_max,
                                                       parent_elt=gr_elt)
            gr_elt.set_content(gr_grp, dtd_min, dtd_max)
        return gr_elt

    def build_sub_elements_dtd_group(self, dtd_group, max_occur,
                                     parent_elt=None):
        """
        Builds a group of the graph that contains XML elements that are
        children of an Xml element (``parent_elt``) and that is defined in
        the DTD with ``dtd_group``.

        This method creates a new group and adds in this group the sub-elements
        declared in ``dtd_group``. Several types of group can exist depending on
        the ``dtd_group`` (sequence, choice) and on the maximum occurence
        of the group.

        Inside a group, we can find sub-groups or sub-elements. This method
        recursively calls the methods of the graph to build the sub-elements
        and the sub-groups.

        :param dtd_group:
            Group definition extracted from the DTD.
        :type dtd_group:
            :class:`~dtd.groups.DtdGroup`
        :param max_occur:
            Maximum occurence of the ``dtd_group`` inside its parent. Can
            be ``INFINITY``.
        :type max_occur:
            :class:`int` or :data:`~xml_struct.utils.INFINITY`
        :param parent_elt:
            Xml element of the graph that will be the parent of the newly
            created group and its children.
        :type parent_elt:
            :class:`~xml_struct.graph_nodes.XMLElement`
        :returns:
            Group that corresponds to the ``dtd_group`` definition in the
            XML Schema and that is created and filled inside this method.
        :rtype:
           :class:`~xml_struct.graph_nodes.Group`
        """
        # Builds the group object that will be inserted in the graph
        if len(dtd_group) == 1:
            # If there is only one child in the group, the group is always a
            # simple group without order.
            gr_grp = Group(parent_elt)
        elif isinstance(dtd_group, DtdSequence):
            gr_grp = Sequence(parent_elt)
        else:
            gr_grp = Alternative(parent_elt)
        # Creates the child elements of the group and adds them in the group
        # object to be inserted in the graph
        for dtd_usg in iter(dtd_group):
            if dtd_usg.max_occur == 0:
                continue
            max_occur =  dtd_usg.max_occur
            if max_occur == dtdConstants.INFINITY:
                max_occur = INFINITY
            if isinstance(dtd_usg.referenced_node, DtdGroup):
                gr_child = self.build_sub_elements_dtd_group(
                    dtd_usg.referenced_node, max_occur,
                    parent_elt=parent_elt)
            else:
                gr_child = self.build_dtd_element(dtd_usg.referenced_node,
                                                  parent_elt=parent_elt)
            gr_grp.add_child(gr_child, dtd_usg.min_occur, max_occur)
        return gr_grp

    def finalize(self):
        # Builds the elements index
        self.elts_index = {}
        for elt in self.elements:
            self.elts_index.setdefault(elt.qname, [])
            self.elts_index[elt.qname].append(elt)
        # Sorts the elements in the graph itself
        self.elements.sort(key=operator.attrgetter("qname"))
        # Finalizes all the elements (simplify, compute_content_type, sort)
        for elt in self.elements:
            elt.finalize()
        # Builds the root elements dictionary for the analysis of the uses
        # inside a library of XML documents
        self.root_uses = {}
        for elt in self.elements:
            if elt.is_high_level:
                self.root_uses[elt] = {}

    def reset_uses(self):
        for elt in self.root_uses:
            self.root_uses[elt] = {}
        for elt in self.elements:
            elt.reset_uses()


TYPES_MAPPING = {
    u"anyType"           : ANY,
    u"anySimpleType"     : STRING,
    u"boolean"           : BOOLEAN,
    u"integer"           : INTEGER,
    u"nonPositiveInteger": INTEGER,
    u"negativeInteger"   : INTEGER,
    u"nonNegativeInteger": INTEGER,
    u"positiveInteger"   : INTEGER,
    u"long"              : INTEGER,
    u"int"               : INTEGER,
    u"short"             : INTEGER,
    u"byte"              : INTEGER,
    u"unsignedLong"      : INTEGER,
    u"unsignedInt"       : INTEGER,
    u"unsignedShort"     : INTEGER,
    u"unsignedByte"      : INTEGER,
    u"decimal"           : FLOAT,
    u"float"             : FLOAT,
    u"double"            : FLOAT,
    u"string"            : STRING,
    u"normalizedString"  : STRING,
    u"Name"              : STRING,
    u"NCName"            : STRING,
    u"QName"             : STRING,
    u"token"             : STRING,
    u"language"          : STRING,
    u"anyURI"            : STRING,
    u"ID"                : STRING,
    u"NOTATION"          : STRING,
    u"NMTOKEN"           : STRING,
    u"IDREF"             : STRING,
    u"ENTITY"            : STRING,
    u"NMTOKENS"          : STRING,
    u"IDREFS"            : STRING,
    u"ENTITIES"          : STRING,
    u"CDATA"             : STRING,
    u"date"              : DATETIME,
    u"time"              : DATETIME,
    u"dateTime"          : DATETIME,
    u"gYear"             : DATETIME,
    u"gYearMonth"        : DATETIME,
    u"gMonth"            : DATETIME,
    u"gMonthDay"         : DATETIME,
    u"gDay"              : DATETIME,
    u"duration"          : DURATION,
    u"base64Binary"      : BASE64,
    u"hexBinary"         : HEXA,
    }
"""
Dictionary that maps the built-in types of XML Schema with elementary
types such as string, boolean, integer, float, datetime, duration,
base64 and hexadecimal.

Type: :class:`dict` of :class:`str` indexed with :class:`unicode`
"""


class MappingTypeMapper(object):
    """Map types according to types in TYPES_MAPPING. Combining returns a single type."""

    @staticmethod
    def map(content_type):
        try:
            return TYPES_MAPPING[content_type]
        except KeyError:
            return UNKNOWN

    @staticmethod
    def combine(type1, type2):
        """
        Tries to find a type that can describe both ``type1`` and ``type2``.

        This function is used when processing a union type (a type that can
        be any of several types) and returns a single type that can be used to
        describe all the types.

        For example, if the first type is a string and the second an integer,
        the function will return a string type. If the first type is an integer
        and the second a float, the function will return a float type, etc.
        In an XML document, all the data can be considered as string data, thus,
        at least, the combination of two different types can be a string type.

        This function uses the constants describing the types that are defined
        in this module.

        :param type1:
            Constant describing the  first type.
        :type type1:
            :class:`unicode`
        :param type2:
            Constant describing the  second type.
        :type type2:
            :class:`unicode`
        :returns:
            Constant describing a type that is combination of the two types
            given in parameter.
        :rtype:
            :class:`unicode`
        """
        if type1 == type2:
            return type1
        if type1 == UNKNOWN:
            return type2
        if type2 == UNKNOWN:
            return type1
        if (type1 == INTEGER and type2 == FLOAT) or \
                (type1 == FLOAT and type2 == INTEGER):
            return FLOAT
        if (type1 == ANY or type2 == ANY):
            return ANY
        return STRING


class NoTypeMapper(object):
    """Alternate type mapper that do no transformation. Combining types may return a set of possible
    types.
    """
    @staticmethod
    def map(content_type):
        return content_type

    @staticmethod
    def combine(type1, type2):
        if not isinstance(type1, set):
            type1 = set([type1])
        if not isinstance(type2, set):
            type2 = set([type2])
        types = type1 | type2
        if len(types) == 1:
            return types.pop()
        return types


DefaultTypeMapper = NoTypeMapper


def compute_xsd_textual_content(xsd_type, type_mapper=DefaultTypeMapper()):
    """
    Extracts information about the textual content of an element or an
    attribute thanks to the ``xsd_type`` object.

    The information consists of the type of the textual content
    converted to an elementary type thanks to
    :data:`~pyxst.xml_struct.graph.TYPES_MAPPING`, a flag
    indicating if the textual content is a space-separated list, and a
    set of predefined values if the textual content can only take
    specified values.

    :param xsd_type:
        XSD type object returns by the schema reader from which the
        information about the textual content will be extracted.
    :type xsd_type:
        :class:`~pyxst.xml_struct.xsd.types.XsdType`
    :returns:
        A tuple composed with the type of the textual content, a flag
        indicating if the textual content is a space-separated list,
        and a set of possible values if the textual content is
        constrained and can only take specified values.
    :rtype:
        Tuple composed with :class:`str`, :class:`bool`, :class:`set`
        of :class:`unicode`
    """
    # Complex type
    if isinstance(xsd_type, XsdComplexType):
        if ( xsd_type.content_type == xsdTypes.ELEMENTS_ONLY
             or xsd_type.content_type == xsdTypes.EMPTY):
            return None, False, None
        if xsd_type.content_type == xsdUtils.MIXED:
            return MIXED, False, None
    # Simple list type
    if xsd_type.content_type == xsdTypes.LIST:
        if isinstance(xsd_type, XsdSimpleListType):
            it_typ, it_lst, it_vals = \
                compute_xsd_textual_content(xsd_type.listitem_type, type_mapper)
            return it_typ, True, it_vals
        else:
            return compute_xsd_textual_content(xsd_type._base_type, type_mapper)
    # Simple union type
    if xsd_type.content_type == xsdTypes.UNION:
        if isinstance(xsd_type, XsdSimpleUnionType):
            typ = UNKNOWN
            lst = True
            vals = set()
            for xsd_sub_type in xsd_type.possible_types:
                sub_typ, sub_lst, sub_vals = \
                    compute_xsd_textual_content(xsd_sub_type, type_mapper)
                typ = type_mapper.combine(typ, sub_typ)
                lst = sub_lst and lst
                if sub_vals is not None and vals is not None:
                    vals = vals.union(sub_vals)
                else:
                    vals = None
            return typ, lst, vals
        else:
            bs_typ, bs_lst, bs_vals = \
                compute_xsd_textual_content(xsd_type._base_type, type_mapper)
            try:
                vals = xsd_type.constraints["enumeration"].value.copy()
                if bs_vals is not None:
                    vals = vals.union(bs_vals)
            except KeyError:
                vals = bs_vals
            return bs_typ, bs_lst, vals
    # Simple type
    typ = type_mapper.map(xsd_type.content_type)
    if xsd_type.content_type in [u"IDREFS", u"ENTITIES", u"NMTOKENS"]:
        lst = True
    else:
        lst = False
    try:
        vals = xsd_type.constraints["enumeration"].value.copy()
    except KeyError:
        vals = None
    return typ, lst, vals


def compute_dtd_textual_content(dtd_content_type):
    """
    Extracts information about the textual content of an element or an
    attribute thanks to the ``dtd_content_type`` value.

    The information consists of the type of the textual content
    converted to an elementary type thanks to
    :data:`~xml_struct.graph.TYPES_MAPPING`, a flag
    indicating if the textual content is a space-separated list.

    :returns:
        A tuple composed with the type of the textual content, a flag
        indicating if the textual content is a space-separated list.
    :rtype:
        Tuple composed with :class:`str`, :class:`bool`
    """
    if ( dtd_content_type == dtdConstants.ELEMENTS_ONLY
         or dtd_content_type == dtdConstants.EMPTY):
        return None, False
    if dtd_content_type == dtdConstants.MIXED:
        return MIXED, False
    if dtd_content_type == dtdConstants.TEXT_ONLY:
        return TYPES_MAPPING[u"string"], False
    try:
        typ = TYPES_MAPPING[dtd_content_type]
    except KeyError:
        typ = UNKNOWN
    if dtd_content_type in [u"IDREFS", u"ENTITIES", u"NMTOKENS"]:
        lst = True
    else:
        lst = False
    return typ, lst




# -*- coding: utf-8 -*-
"""
This module defines the constants and classes used to represent the nodes
read in an XML Schema.

Constants
.........

.. autodata:: ATTRIBUTE

.. autodata:: ATTRIBUTE_GROUP

.. autodata:: ELEMENT

.. autodata:: GROUP

.. autodata:: TYPE

.. autodata:: XSD_NS

Nodes
.....

.. autoclass:: XsdNode
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdUsage
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"


from .errors import XsdError


ATTRIBUTE = "Attribute"
"""
Constant containing the category of the attribute nodes

Type: :class:`str`
"""

ATTRIBUTE_GROUP = "Attribute Group"
"""
Constant containing the category of the attribute group nodes

Type: :class:`str`
"""

ELEMENT = "Element"
"""
Constant containing the category of the element nodes

Type: :class:`str`
"""

TYPE = "Type"
"""
Constant containing the category of the type nodes

Type: :class:`str`
"""

GROUP = "Group"
"""
Constant containing the category of the group nodes

Type: :class:`str`
"""

INFINITY = "infinity"
"""
Constant describing the infinity (for unbounded occurences).

Type: :class:`str`
"""

MIXED = "Mixed Content"
"""
Constant describing the type of the content that can occur inside an element
if this element can contain, at the same time, sub-elements and textual data.

Type: :class:`str`
"""

XSD_NS = u"http://www.w3.org/2001/XMLSchema"
"""
XSD namespace as defined in the XML Schema standard.

Type: :class:`unicode`
"""


def split_qname(qname):
    """
    Splits the XML qualified name in a tuple containing the namespace and the
    local name.

    A qualified name is composed with the namespace and the local name.
    In the XML nodes read with :mod:`xml.etree` module, the qualified name
    is built as follow: ``{namespace}local-name``.

    :param qname:
        XML qualified name as built by the :mod:`xml.etree` module.
    :type qname:
        :class:`unicode`
    :returns:
        The namespace and the local name extracted from ``qname``
    :rtype:
        tuple composed with :class:`unicode` and :class:`unicode`
    """
    if qname is None:
        return None, None
    name = str(qname)
    if name.find('}') == -1:
        return None, name
    return name[name.find('{')+1:name.find('}')], name[name.find('}')+1:]


def resolve_xml_qname(qname, ns_decl):
    """
    Resolves the XML qualified name and returns this qualified name as if it
    was read by the :mod:`xml.etree` module.

    A qualified name is composed with the namespace and the local name.
    In the XML document, the qualified name is built as follow:
    ``prefix:local-name`` and namespace declarations map the prefixes with
    the actual namespaces.

    :param qname:
        XML qualified name as read in the XML document.
    :type qname:
        :class:`unicode`
    :param ns_decl:
        Mapping between the prefixes used in the XML document and the
        namespaces.
    :type ns_decl:
        Dictionary of :class:`unicode` indexed with :class:`unicode`.
    :returns:
        The qualified name structured as built by the :mod:`xml.etree`
        module (``{namespace}local-name``)
    :rtype:
        :class:`unicode`
    """
    if qname is None:
        return None
    name = str(qname)
    if name.find(':') == -1:
        prefix = ""
        local_name = name
    else:
        prefix = name[:name.find(':')]
        local_name = name[name.find(':')+1:]
    if prefix == "" and "" in ns_decl:
        namespace = ns_decl[u""]
    elif prefix == "" and None in ns_decl:
        namespace = ns_decl[None]
    elif prefix == "":
        return local_name
    elif prefix == u"xml":
        namespace = "http://www.w3.org/XML/1998/namespace"
    else:
        try:
            namespace = ns_decl[prefix]
        except KeyError:
            raise Exception("%s prefix is not bound to a namespace" % prefix)
    return "{%s}%s" % (namespace, local_name)

class XsdNode(object):
    """
    Base class representing any node in an XML Schema.

    .. attribute:: category

       Class attribute representing the category of the node (element,
       attribute, type, group, etc.)

       Type: :class:`str`

    .. attribute:: name

       Name of the node. If the node in the XML Schema doesn't have an
       explicit name, an implicit name is created while parsing the schema.

       Type: :class:`unicode`

    .. attribute:: uid

       Unique identifier of the node in its category.

       Type: :class:`unicode`

    .. attribute:: desc

       List of the documentation data given inside the schema for this node.

       Type: :class:`list` of :class:`unicode`

    .. automethod:: __init__
    """
    category = ""

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new node.

        :param name:
            Name of the node.
        :type name:
            :class:`unicode`
        """
        self.name = name
        self.uid_prefix = uid_prefix
        self.desc = []
        if self.__class__ is XsdNode:
            raise NotImplementedError("Can't instanciate abstract class")

    @property
    def uid(self):
        return u"{0}/{1}".format(self.uid_prefix, self.name)

    def __repr__(self):
        """
        Returns a string representation of the node.
        """
        return u"<{0} {1}>".format(self.category, self.name)


class XsdUsage(XsdNode):
    """
    Class representing in an XML Schema the usage of another node.

    Usages are typically inserted inside a group in a complex element to
    point sub-elements or sub-groups and indicate the minimum and maximum
    occurences (e.g. ``<xsd:element ref="sub-elt" minOccurs="0"/>``).

    If a sub-element (an attribute or a sub-group) is directly defined
    (```<xsd:element name="sub-elt" maxOccurs="4"/>``), a usage is
    nevertheless inserted between the group and the sub-element to
    indicate the minimum and maximum occurences of this sub-element.

    .. attribute:: category

       Category (element, attribute, group) of the referenced node.

       Type: :class:`str`

    .. attribute:: min_occur

       Minimum occurence of the referenced node in the context where the usage
       is inserted.

       Type: :class:`int`

    .. attribute:: max_occur

       Maximum occurence of the referenced node in the context where the usage
       is inserted.

       Type: :class:`int`

    .. attribute:: referenced_node

       Node referenced by this usage.

       Type: :class:`~pyxst.xml_struct.xsd.utils.XsdNode`

    .. automethod:: __init__
    """
    def __init__(self, ref_node, min_occur=1, max_occur=1):
        """
        Initializes a new usage for ``ref_node``.

        :param ref_node:
            Referenced node used by this usage.
        :type ref_node:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement` or
            :class:`~pyxst.xml_struct.xsd.attributes.nodes.XsdAttribute`
            or :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        :param min_occur:
            Minimum occurence for the referenced node in the
            context of this usage.
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence for the referenced node in the
            context of this usage.
        :type max_occur:
            :class:`int` or
            :data:`~pyxst.xml_struct.utils.INFINITY`
        """
        if ref_node.category not in [ATTRIBUTE, ELEMENT, GROUP]:
            raise XsdError("%s cannot be referenced in group or complex type "
                           "(only attributes, elements or groups can)"
                           % ref_node)
        if min_occur < 0:
            raise XsdError("Minimum occurence for %s must be positive or null"
                           % ref_node)
        if not max_occur is INFINITY and max_occur < min_occur:
            raise XsdError("Maximum occurence for %s must be greater than "
                           "minimum occurence" % ref_node)
        super(XsdUsage, self).__init__(ref_node.name)
        self.min_occur = min_occur
        self.max_occur = max_occur
        self.referenced_node = ref_node

    @property
    def category(self):
        return self.referenced_node.category

    @property
    def uid(self):
        return self.referenced_node.uid

    def __repr__(self):
        return ("<Usage (%s,%s) of %s %s>"
                % (self.min_occur, self.max_occur, self.category, self.name))



# -*- coding: utf-8 -*-


from .errors import DtdError, XsdError
from .constants import INFINITY, ATTRIBUTE, ELEMENT, GROUP


def resolve_xml_qname(qname, ns_decl):
    """
    Resolves the XML qualified name and returns this qualified name as if it
    was read by the :mod:`xml.etree` module.

    A qualified name is composed with the namespace and the local name.
    In the XML document, the qualified name is built as follow: 
    ``prefix:local-name`` and namespace declarations map the prefixes with
    the actual namespaces.

    :param qname:
        XML qualified name as read in the XML document.
    :type qname:
        :class:`unicode`
    :param ns_decl:
        Mapping between the prefixes used in the XML document and the 
        namespaces.
    :type ns_decl:
        Dictionary of :class:`unicode` indexed with :class:`unicode`.
    :returns:
        The qualified name structured as built by the :mod:`xml.etree`
        module (``{namespace}local-name``)
    :rtype:
        :class:`unicode`
    """
    if qname is None:
        return None
    name = str(qname)
    if name.find(':') == -1:
        prefix = ""
        local_name = name
    else:
        prefix = name[:name.find(':')]
        local_name = name[name.find(':')+1:]
    if prefix == "" and "" in ns_decl:
        namespace = ns_decl[u""]
    elif prefix == "" and None in ns_decl:
        namespace = ns_decl[None]
    elif prefix == "":
        return local_name
    elif prefix == u"xml":
        namespace = "http://www.w3.org/XML/1998/namespace"
    else:
        try:
            namespace = ns_decl[prefix]
        except KeyError:
            raise Exception("%s prefix is not bound to a namespace" % prefix)
    return "{%s}%s" % (namespace, local_name)


class DtdNode(object):
    """
    Base class representing any node in a DTD.

    .. attribute:: category

       Class attribute representing the category of the node (element,
       attribute, type, group, etc.)

       Type: :class:`unicode`

    .. attribute:: name

       Name of the node. If the node in the DTD doesn't have an
       explicit name, an implicit name is created while parsing the DTD.

       Type: :class:`unicode`

    .. attribute:: uid

       Unique identifier of the node in its category.

       Type: :class:`unicode`
    """
    category = u""

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new node.

        :param name:
            Name of the node.
        :type name:
            :class:`unicode`
        """
        self.name = name
        self.uid_prefix = uid_prefix
        if self.__class__ is DtdNode:
            raise NotImplementedError("Can't instanciate abstract class")

    @property
    def uid(self):
        return u"{0}/{1}".format(self.uid_prefix, self.name)

    def __repr__(self):
        """
        Returns a string representation of the node.
        """
        return u"<{0} {1}>".format(self.category, self.name)


class DtdUsage(DtdNode):
    """
    Class representing in a DTD the usage of another node

    Usages are typically inserted inside a group in the sub-content of an
    element to point sub-elements or sub-groups and indicate the minimum and
    maximum occurences (e.g. ``(sub-elt)+``).

    .. attribute:: category

       Category (element, attribute, group) of the referenced node.

       Type: :class:`unicode`

    .. attribute:: min_occur

       Minimum occurence of the referenced node in the context where the usage
       is inserted.

       Type: :class:`int`

    .. attribute:: max_occur

       Maximum occurence of the referenced node in the context where the usage
       is inserted. Can be ``INFINITE``.

       Type: :class:`int`

    .. attribute:: referenced_node

       Node referenced by this usage.

       Type: :class:`~dtd.utils.DtdNode`

    .. automethod:: __init__
    """
    def __init__(self, ref_node, min_occur=1, max_occur=1):
        """
        Initializes a new usage for ``ref_node``.
        """
        if ref_node.category not in [ATTRIBUTE, ELEMENT, GROUP]:
            raise DtdError("%s cannot be referenced in group or complex type "
                           "(only attributes, elements or groups can)"
                           % ref_node)
        if min_occur < 0:
            raise XsdError("Minimum occurence for %s must be positive or null"
                           % ref_node)
        if not max_occur is INFINITY and max_occur < min_occur:
            raise XsdError("Maximum occurence for %s must be greater than "
                           "minimum occurence" % ref_node)
        super(DtdUsage, self).__init__(ref_node.name)
        self.min_occur = min_occur
        self.max_occur = max_occur
        self.referenced_node = ref_node

    @property
    def category(self):
        return self.referenced_node.category

    @property
    def uid(self):
        return self.referenced_node.uid

    def __repr__(self):
        return ("<Usage (%s,%s) of %s %s>"
                % (self.min_occur, self.max_occur, self.category, self.name))



import sys


class PrettyPrinter(object):
    def __init__(self, print_qname=True, indent='  '):
        self.print_qname = print_qname
        self.indent = indent

    def pprint(self, element, out=sys.stdout, indent=''):
        callback = getattr(self, 'visit_' + element.__class__.__name__.lower())
        callback(out, element, indent)

    def visit_xmlelement(self, out, element, indent):
        if self.print_qname:
            name = element.local_name
        else:
            name = "{%s}%s" % (element.namespace, element.local_name)
        out.write(indent + '<' + name + '>' + '\n')
        if element.content:
            self.pprint(element.content, out, indent + self.indent)

    def visit_occurence(self, out, element, indent):
        if element.minimum == element.maximum:
            occurs = u"[{0}]".format(element.minimum)
        else:
            occurs = u"[{0},{1}]".format(element.minimum, element.maximum)
        out.write(indent + occurs + '\n')
        self.pprint(element.target, out, indent + self.indent)

    def visit_group(self, out, element, indent):
        out.write(indent + element.__class__.__name__ + '\n')
        for occ in element.children:
            self.pprint(occ, out, indent + self.indent)

    visit_alternative = visit_sequence = visit_group



# -*- coding: utf-8 -*-


from .errors import DtdError
from .utils import DtdNode, DtdUsage
from .constants import ELEMENT, ATTRIBUTE


class DtdElement(DtdNode):
    """
    Class representing an element defined in a DTD.
    """
    category = ELEMENT

    def __init__(self, name, uid_prefix=u""):
        super(DtdElement, self).__init__(name, uid_prefix=uid_prefix)
        self.attributes = {}
        self.content_type = None
        self.sub_content = None

    def add_attribute(self, attr, min_occur=1, max_occur=1):
        """
        Adds the ``attr`` attribute to the element.

        This method checks that ``attr`` category is ``ATTRIBUTE``, that
        max_occur is never greater than 1 and that there is not already an
        attribute with the same identifier in this element.
        """
        if not attr.category is ATTRIBUTE:
            raise DtdError(u"Can't add {0} to {1} as attribute"
                           u"".format(attr, self))
        if attr.uid in self.attributes:
            raise DtdError(u"{0} already declared in {1}".format(attr, self))
        if max_occur > 1:
            raise DtdError(u"{0} can't occur more than one time in {1}"
                           u"".format(attr, self))
        self.attributes[attr.uid] = DtdUsage(attr, min_occur, max_occur)


class DtdAttribute(DtdNode):
    """
    Class representing an attribute defined in an XML Schema.
    """
    category = ATTRIBUTE

    def __init__(self, name, uid_prefix=u""):
        super(DtdAttribute, self).__init__(name, uid_prefix=uid_prefix)
        self.dtd_type = None
        self.fixed_value = False
        self.default_value = None
        self.possible_values = None

    @property
    def uid(self):
        return u"{0}/@{1}".format(self.uid_prefix, self.name)



# -*- coding: utf-8 -*-
"""
This module defines the classes that represent the attributes or the
groups of attributes read in an XML Schema.

The attributes and the groups of attributes are defined as a specific
nodes derived from
:class:`~pyxst.xml_struct.xsd.utils.XsdNode`.

.. autoclass:: XsdAttribute
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdAttributeGroup
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"


from .errors import XsdError
from .utils import ATTRIBUTE, ATTRIBUTE_GROUP
from .utils import XsdNode


class XsdAttribute(XsdNode):
    """
    Class representing an attribute defined in an XML Schema.

    .. attribute:: xsd_type

       Type of this attribute declared in the XML Schema. This type is a
       type node.

       Type: :class:`~pyxst.xml_struct.xsd.types.XsdType`

    For this class, the ``category`` class attribute is set to ``ATTRIBUTE``.

    .. automethod:: __init__
    """
    category = ATTRIBUTE

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new attribute node.

        :param name:
            Name of the attribute.
        :type name:
            :class:`unicode`
        """
        super(XsdAttribute, self).__init__(name, uid_prefix=uid_prefix)
        self.xsd_type = None

    @property
    def uid(self):
        return u"{0}/@{1}".format(self.uid_prefix, self.name)

class XsdAttributeGroup(XsdNode):
    """
    Class representing an attribute group defined in an XML Schema.

    Attribute groups are only defined to group attributes and ease the
    elements definition.

    As attribute groups can contain attribute groups, before iterating over
    the attributes of this group, it is necessary to collect all the attributes
    defined in the sub-groups.

    .. attribute:: attribute_uses

       List of attribute uses defined in the attribute group.

       Each attribute use is kept as a tuple composed with the attribute
       node itself, the minimum occurence and the maximum occurence.

       Type: list of tuples
       (:class:`~pyxst.xml_struct.xsd_nodes.XsdAttribute`,
       :class:`int`, :class:`int`)

    .. attribute:: attr_groups

       List of attribute groups used in this attribute group.

       Type: list of
       :class:`~pyxst.xml_struct.xsd_nodes.XsdAttributeGroup`

    For this class, the ``category`` class attribute is set to
    ``ATTRIBUTE_GROUP``.

    .. automethod:: __init__
    .. automethod:: __iter__
    """
    category = ATTRIBUTE_GROUP

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new attribute group node.

        :param name:
            Name of the attribute group.
        :type name:
            :class:`unicode`
        """
        super(XsdAttributeGroup, self).__init__(name, uid_prefix=uid_prefix)
        self.attribute_uses = []
        self.attr_groups = []

    def add_attribute(self, attr, min_occur=1, max_occur=1):
        """
        Adds the ``attr`` attribute to the attribute group.

        This method only checks that ``attr`` category is ``ATTRIBUTE``. The
        other checks will be performed when the attribute will be inserted in
        the complex types where the attribute group is inserted.

        :param attr:
            Attribute node to be added in this element.
        :type attr:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`
        :param min_occur:
            Minimum occurence for the attribute (0 or 1)
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence for the attribute (0 or 1)
        :type max_occur:
            :class:`int`
        """
        if not attr.category is ATTRIBUTE:
            raise XsdError("Can't add %s to %s as attribute" % (attr, self))
        self.attribute_uses.append( (attr, min_occur, max_occur) )

    def add_attribute_group(self, attr_grp):
        """
        Adds the ``attr_grp`` attribute group to this attribute group.

        This method only checks ``attr_grp`` category is ``ATTRIBUTE_GROUP``.

        :param attr_grp:
            Attribute group node to be added in this attribute group.
        :type attr_grp:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`
        """
        if not attr_grp.category is ATTRIBUTE_GROUP:
            raise XsdError("Can't add %s to %s as attribute group"
                           % (attr_grp, self))
        self.attr_groups.append(attr_grp)

    def __iter__(self):
        """
        Returns an iterator over the attribute uses defined in this
        attribute group and in all its sub-groups.

        The items returned by the iterator are tuples composed with
        (:class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`,
        :class:`int`, :class:`int`)

        :returns:
            An iterator over all the attribute uses defined in this attribute
            group and its sub-groups.
        :rtype:
            :class:`listiterator`
        """
        lst = self.attribute_uses[:]
        for grp in self.attr_groups:
            for attr_use in iter(grp):
                lst.append(attr_use)
        return iter(lst)

    @property
    def uid(self):
        return u"{0}/@@{1}".format(self.uid_prefix, self.name)



# -*- coding: utf-8 -*-
"""
This module defines the classes that represent the constraints that can
be defined in an XML Schema when restricting a simple type.

.. autoclass:: XsdConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdMinConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdStrictMinConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdMaxConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdStrictMaxConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdMaxDigitsConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdMinDigitsConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdLengthConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdMaxLengthConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdMinLengthConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdPossibleValuesConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdValuePatternConstraint
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XsdWhiteSpacePolicyConstraint
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"


from .errors import XsdError


class XsdConstraint(object):
    """
    Base class for the constraints defined in the restriction of the simple
    types.

    In an XML Schema, several sorts of constraint can be defined:
    ``minExclusive``,``minInclusive``, ``maxExclusive``,
    ``maxInclusive``, ``totalDigits``, ``fractionDigits``, ``length``,
    ``minLength``, ``maxLength``, ``enumeration``, ``whiteSpace`` and
    ``pattern``.

    .. attribute:: value

       Value associated to the constraint. The exact type of the value
       depends on the sort of constraint (minimum, policy, set of possible
       values). See the derived classes for more details.

    .. attribute:: constraint_phrase

       Class attribute defining a little phrase that explains the
       constraint (used for the string representation).

       Type: :class:`str`
    """
    constraint_phrase = ""

    def __init__(self):
        """
        Initializes a new constraint.
        """
        self.value = None
        if self.__class__ is XsdConstraint:
            raise NotImplementedError("Abstract classes can't be implemented")

    def define_value(self, value):
        """
        Defines the value associated with the constraint.

        If the new value is less constraining than the
        value actually defined in the constraint, the new value is not
        taken into account.

        This abstract method will be defined in derived classes.

        :param value:
            Value associated with the constraint.
        :type value:
            Depends on the actual constraint (see derived classes).
        """
        raise NotImplementedError("Abstract method can't be called")

    def __repr__(self):
        """
        Returns a string representation of the constraint.

        :returns:
            String representation of the constraint
        :rtype:
            :class:`str`
        """
        return "%s %s" % (self.constraint_phrase, self.value)


class XsdMinConstraint(XsdConstraint):
    """
    Constraint corresponding to ``minInclusive`` that gives the minimum of
    the content found in a simple type.

    The value associated to this constraint is a :class:`float`.
    """
    constraint_phrase = "Value >= "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (minimum).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`float`
        """
        try:
            value = float(value)
        except ValueError:
            raise XsdError("Value \"%s\" in constraint should be a number"
                           % value)
        if self.value is None or self.value < value:
            self.value = value


class XsdStrictMinConstraint(XsdConstraint):
    """
    Constraint corresponding to ``minExclusive`` that gives the strict
    minimum of the content found in a simple type (content can't be
    equal to the minimum).

    The value associated to this constraint is a :class:`float`.
    """
    constraint_phrase = "Value > "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (strict minimum).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`float`
        """
        try:
            value = float(value)
        except ValueError:
            raise XsdError("Value \"%s\" in constraint should be a number"
                           % value)
        if self.value is None or self.value < value:
            self.value = value


class XsdMaxConstraint(XsdConstraint):
    """
    Constraint corresponding to ``maxInclusive`` that gives the maximum of
    the content found in a simple type.

    The value associated to this constraint is a :class:`float`.
    """
    constraint_phrase = "Value <= "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (maximum).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`float`
        """
        try:
            value = float(value)
        except ValueError:
            raise XsdError("Value \"%s\" in constraint should be a number"
                           % value)
        if self.value is None or self.value > value:
            self.value = value


class XsdStrictMaxConstraint(XsdConstraint):
    """
    Constraint corresponding to ``maxExclusive`` that gives the strict
    maximum of the content found in a simple type (content can't be
    equal to maximum).

    The value associated to this constraint is a :class:`float`.
    """
    constraint_phrase = "Value < "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (strict maximum).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`float`
        """
        try:
            value = float(value)
        except ValueError:
            raise XsdError("Value \"%s\" in constraint should be a number"
                           % value)
        if self.value is None or self.value > value:
            self.value = value


class XsdMaxDigitsConstraint(XsdConstraint):
    """
    Constraint corresponding to ``totalDigits`` that gives the maximum number
    of digits of the content found in simple type.

    The value associated to this constraint is an :class:`int`.
    """
    constraint_phrase = "Number of digits of Value <= "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (maximum number of
        digits).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`int`
        """
        try:
            value = int(value)
        except ValueError:
            raise XsdError("Value %s in constraint should be an integer"
                           % value)
        if self.value is None or self.value > value:
            self.value = value


class XsdMinDigitsConstraint(XsdConstraint):
    """
    Constraint corresponding to ``fractionDigits`` that gives the
    minimum number of digits of the content found in simple type.

    The value associated to this constraint is an :class:`int`.
    """
    constraint_phrase = "Number of digits of Value >= "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (minimum number of
        digits).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`int`
        """
        try:
            value = int(value)
        except ValueError:
            raise XsdError("Value %s in constraint should be an integer"
                           % value)
        if self.value is None or self.value < value:
            self.value = value


class XsdLengthConstraint(XsdConstraint):
    """
    Constraint corresponding to ``length`` that gives the length of the
    content found in simple type (string length or list length).

    The value associated to this constraint is an :class:`int`.
    """
    constraint_phrase = "Length of Value == "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (length).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`int`
        """
        try:
            value = int(value)
        except ValueError:
            raise XsdError("Value %s in constraint should be an integer"
                           % value)
        self.value = value


class XsdMaxLengthConstraint(XsdConstraint):
    """
    Constraint corresponding to ``maxLength`` that gives the maximum
    length of the content found in simple type (string length or list
    length).

    The value associated to this constraint is an :class:`int`.
    """
    constraint_phrase = "Length of Value <= "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (maximum length).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`int`
        """
        try:
            value = int(value)
        except ValueError:
            raise XsdError("Value %s in constraint should be an integer"
                           % value)
        if self.value is None or self.value > value:
            self.value = value


class XsdMinLengthConstraint(XsdConstraint):
    """
    Constraint corresponding to  ``minLength`` that gives the minimum
    length of the content found in simple type (string length or list
    length).

    The value associated to this constraint is an :class:`int`.
    """
    constraint_phrase = "Length of Value >= "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (minimum length).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`int`
        """
        try:
            value = int(value)
        except ValueError:
            raise XsdError("Value %s in constraint should be an integer"
                           % value)
        if self.value is None or self.value < value:
            self.value = value


class XsdPossibleValuesConstraint(XsdConstraint):
    """
    Constraint corresponding to ``enumeration`` that gives the possible
    values of the content found in the simple type.

    The value associated to this constraint is a :class:`set` of
    :class:`unicode` that gives all the possible values for the
    content.
    """
    constraint_phrase = "Value in "

    def define_value(self, value):
        """
        Defines a new possible value for the constraint (defined set
        of possible values).

        :param value:
            New value associated with the constraint.
        :type value:
            :class:`unicode`
        """
        if self.value is None:
            self.value = set()
        self.value.add(value)


class XsdValuePatternConstraint(XsdConstraint):
    """
    Constraint corresponding to ``pattern`` that gives the pattern
    expected for the content of simple type. The pattern is defined as
    a regular expression.

    The value associated to this constraint is a :class:`unicode`.
    """
    constraint_phrase = "Value matches RegExp: "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (pattern of
        regular expression).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`unicode`
        """
        self.value = value


class XsdWhiteSpacePolicyConstraint(XsdConstraint):
    """
    Constraint corresponding to ``whiteSpace`` that gives the policy
    for processing the white spaces in the content found in the simple
    type.

    The value associated to this constraint is a :class:`unicode`.
    """
    constraint_phrase = "White spaces processing in Value: "

    def define_value(self, value):
        """
        Defines the value associated with the constraint (chosen
        policy for the white spaces processing).

        :param value:
            Value associated with the constraint.
        :type value:
            :class:`unicode`
        """
        self.value = value



# -*- coding: utf-8 -*-
"""
This module contains the classes that represent relations between XML
elements of the graph that can be translated in some arKItect
concepts. It also contains the function that creates these relations in
the graph decribing the XML structure by analyzing the structure of XML
elements.

The graph that describes the XML structure is a
:class:`pyxst.xml_struct.graph.XMLGraph` object. It
contains several nodes including XML elements that come from the
:mod:`pyxst.xml_struct.graph_nodes` module.

.. autoclass:: TargetRelation
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: IsParentOf
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: HasAttribute
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: HasListAttribute
   :members:
   :undoc-members:
   :show-inheritance: 

.. autodata:: RELATION_CLASS

.. autodata:: IS_PARENT_OF
.. autodata:: HAS_ATTRIBUTE
.. autodata:: HAS_LIST_ATTRIBUTE

.. autofunction:: weave_target_relations

.. autofunction:: build_relations_in_group
"""
__docformat__ = "restructuredtext en"


from .utils import MIXED, INFINITY
from .graph_nodes import Group


class TargetRelation(object):
    """
    Abstract class that describes a target relation between two nodes
    of the graph (XML elements or XML attributes).

    .. attribute:: start

       Node where the relation starts from.

       Type: :class:`~pyxst.xml_struct.graph_nodes.XMLNode`

    .. attribute:: end

       Node where the relation ends to.

       Type: :class:`~pyxst.xml_struct.graph_nodes.XMLNode`
 
    .. attribute:: DESC

       Class attribute containing the textual description of the relation 
       that will be used to display the text representation of this object.

       Type: :class:`str`
    """
    DESC = "%s is in relation with %s"

    def __init__(self, start_node, end_node):
        """
        Initializes a new target relation between two graph nodes.

        :param start_node:
            Node where the relation starts from.
        :type start_node:
            :class:`~pyxst.xml_struct.graph_nodes.XMLNode`
        :param end_node:
            Node where the relation ends to.
        :type end_node:
            :class:`~pyxst.xml_struct.graph_nodes.XMLNode`
        """
        self.start = start_node
        self.end = end_node
        if self.__class__ is TargetRelation:
            raise NotImplementedError("Abstract classes can't be instantiated")

    def __repr__(self):
        """
        Returns a string representation of the object.

        Uses ``DESC`` class attribute to build this representation.

        :returns:
            Text representation of the object.
        :rtype:
            :class:`str`
        """
        return self.DESC % (self.start, self.end)


class IsParentOf(TargetRelation):
    """
    Target relation that describes an XML element will be the parent of another
    XML element in the target structure in arKItect.
    """
    DESC = "%s is parent of %s"


class HasAttribute(TargetRelation):
    """
    Target relation that describes an XML element will have an XML element
    or an XML attribute as an attribute in the target structure in arKItect.
    """
    DESC = "%s has %s as attribute"


class HasListAttribute(TargetRelation):
    DESC = "%s has %s as list attribute"


IS_PARENT_OF = "IsParentOf Relation"
"""
Constant describing the "IsParentOf" relation.

Type: :class:`str`
"""

HAS_ATTRIBUTE = "HasAttribute Relation"
"""
Constant describing the "HasAttribute" relation.

Type: :class:`str`
"""

HAS_LIST_ATTRIBUTE = "HasListAttribute Relation"
"""
Constant describing the "HasListAttribute" relation.

Type: :class:`str`
"""

RELATION_CLASS = {IS_PARENT_OF: IsParentOf,
                  HAS_ATTRIBUTE: HasAttribute,
                  HAS_LIST_ATTRIBUTE: HasListAttribute,
                  }
"""
Dictionary mapping the names of the target relations with the classes that
represent them.

Type: :class:`dict` of :class:`class` indexed with :class:`str`
"""


def weave_target_relations(graph):
    """
    Function that creates target relations between the nodes of the graph
    that describes the XML structure.

    Target relations describe relations that can be translated into
    arKItect concepts:

    * "has attribute" between an XML element and an XML attribute or
      between an XML element and an XML sub-element that contains
      textual data and no attribute;

    * "has list attribute" between an XML element and an XML attribute
      that contains a list of values or between an XML element and an
      XML sub-element that contains textual data and no attribute but
      can occur several times or between an XML element and an XML
      sub-element that contains a textual list of values and no attribute;

    * "is parent of" between an XML element and an XML sub-element that 
      contains mixed content or sub-elements.

    :param graph:
        Graph that decribes the XML structure. The target relations will be
        created inside this graph between the graph nodes.
    :type graph:
        :class:`~pyxst.xml_struct.graph.XMLGraph`
    """
    for elt in graph.elements:
        for occur in elt.attributes.values():
            attr = occur.target
            if attr.textual_content_is_list:
                graph.add_target_relation(HAS_LIST_ATTRIBUTE, elt, attr)
            else:
                graph.add_target_relation(HAS_ATTRIBUTE, elt, attr)
        if elt.content is not None:
            grp = elt.content.target
            grp_max = elt.content.maximum
            build_relations_in_group(graph, elt, grp, grp_max)


def build_relations_in_group(graph, parent_elt, grp, grp_max):
    """
    Private function that builds the target relations between a parent element
    and the sub-elements that occur inside a group.

    As groups can contain sub-groups, this function can be recursively called
    in order to completely process the sub-content of the parent element.

    This function should not be directly called. Use ``weave_target_relations``
    function instead.

    :param graph:
        Graph where the target relations are weaved in.
    :type graph:
        :class:`~pyxst.xml_struct.graph.XMLGraph`
    :param parent_elt:
        XML Element where the considered group is defined. This element will
        be the starting point of the target relations.
    :type parent_elt:
        :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
    :param grp:
        Group that contains the sub-elements of ``parent_elt``. This group
        might also contain sub-groups that contain the sub-elements.
    :type grp:
        :class:`~pyxst.xml_struct.graph_nodes.Group`
    :param grp_max:
        Maximum occurence of the ``grp`` group. This maximum occurence has been
        cumulated since the reading of ``parent_elt`` content.
    :type grp_max:
        :class:`int` or :data:`~pyxst.xml_struct.utils.INFINITY`
    """
    for occur in grp.children:
        sub_item = occur.target
        if occur.maximum == INFINITY or grp_max == INFINITY:
            sub_max = INFINITY
        else:
            sub_max = occur.maximum * grp_max
        if isinstance(sub_item, Group):
            build_relations_in_group(graph, parent_elt, sub_item, sub_max)
        else:
            if sub_item.textual_content_type in [None, MIXED] or \
                    len(sub_item.attributes) > 0:
                graph.add_target_relation(IS_PARENT_OF, parent_elt, sub_item)
            elif sub_item.textual_content_is_list:
                graph.add_target_relation(HAS_LIST_ATTRIBUTE,
                                          parent_elt, sub_item)
            elif sub_max == INFINITY or sub_max > 1:
                graph.add_target_relation(HAS_LIST_ATTRIBUTE,
                                          parent_elt, sub_item)
            else:
                graph.add_target_relation(HAS_ATTRIBUTE, parent_elt, sub_item)
